"""
The module ``snewpy.snowglobes_interface`` contains a low-level Python interface for SNOwGLoBES v1.3.

.. note::
    Users should only use the high-level interface described above.
    This low-level interface is not guaranteed to be stable and may change at
    any time without warning, e.g. to support new SNOwGLoBES versions.
"""
from pathlib import Path
import pandas as pd
import numpy as np
import os

import logging
logger = logging.getLogger(__name__)

from tqdm.auto import tqdm

import snowglobes_data
import sys
if sys.version_info >= (3, 9):
    from importlib.resources import files
else:
    from importlib_resources import files

def guess_material(detector):
    if detector.startswith(('wc', 'ice', 'km3net')):
        mat = 'water'
    elif detector.startswith('d2O'):
        mat = 'heavywater'
    elif detector.startswith('ar'):
        mat = 'argon'
    elif detector.startswith('nova'):
        mat = 'nova_soup'
    elif detector.startswith('halo'):
        mat = 'lead'
    elif detector.startswith('scint'):
        mat = 'scint'
    else: 
        raise ValueError(f'Please provide material for {detector}')

    if '_he' in detector:
        mat = mat+'_he'
    return mat


class SimpleRate():
    def __init__(self, detectors:str="all", detector_effects=True, base_dir:Path=''):
        """Simple rate calculation interface.
        Computes expected rate for a detector without using GLOBES. The formula for the rate is

                Rate = Sum_i ([cross-section in 10^-38 cm^2] x 10^-38 x [fluence in cm^-2])_i x [smearing matrix]_{ij}
                       x [target mass in kton] x [Dalton per kton] x [energy bin size in GeV] x [efficiency]

        with [target mass in kton] x [Dalton per kton] = number of reference targets in experiment.

        On construction the code will read: 

        * detectors from `<base_dir>/detector_configurations.dat`,
        * channels from `<base_dir>/channels/channel_*.dat`

        if the detector_effects option is on, also read efficiencies and smearing:

        * efficiencies from `<base_dir>/effic/effic_*.dat`,
        * smearing matrices from `<base_dir>/smear/smear_*.dat`

        After that use :meth:`SimpleRate.run` method to run the simulation for specific detector and flux file.

        Parameters
        ----------
        detectors: str
            Name of detector. If ``"all"``, will use all detectors supported by SNOwGLoBES.

        detector_effects: bool
            If true, account for efficiency and smearing. If false, consider a perfect detector.

        base_dir:         Path or None
            Path to the directory where the cross-section, detector, and channel files are located.
            If empty, try to get it from ``$SNOWGLOBES`` environment var or the `snowglobes_data` module.
        """
        try:
            self.base_dir = Path(base_dir if base_dir else os.environ['SNOWGLOBES'])
        except KeyError:
            print("Using snowglobes_data module ...")
            self.base_dir = files(snowglobes_data)
        self._load_detectors(self.base_dir/'detector_configurations.dat', detectors)
        self._load_channels(self.base_dir/'channels')
        self.efficiencies = None
        self.smearings = None
        if detector_effects:
            self._load_efficiency_vectors(self.base_dir/'effic')
            self._load_smearing_matrices(self.base_dir/'smear')

    def _load_detectors(self, path:Path, detectors:str):
        df = pd.read_table(path,names=['name','mass','factor'], delim_whitespace=True, comment='#')
        df['tgt_mass']=df.mass*df.factor

        if detectors == 'all':
            detectors = list(df['name'])
            detectors.remove('d2O')
        elif isinstance(detectors, str):
            detectors = [detectors]

        self.detectors = df.set_index('name').T.filter(items=detectors)
        logger.info(f'read masses for detectors {list(self.detectors)}')
        logger.debug(f'detectors: {self.detectors}')
       
    def _load_channels(self, chan_dir):
        self.channels = {}
        self.binning = {}
        for f in chan_dir.glob('channels_*.dat'):
            material = f.stem[len('channels_'):]
            # in SNOwGLoBES v1.3, there are three different formats of channel files
            # We identify these based on their first line and read them in separately:
            with open(f) as _f:
                l = _f.readline().strip()
            if l.startswith('%'):
                # Format 1: explicit binning, same for all channels
                tokens = l.split()[1:]
                df = pd.read_table(f, delim_whitespace=True, index_col=1, comment='%', names=['name','n','parity','flavor','weight'])
            elif l.startswith('SN_nu'):
                # Format 2: explicit binning, differs per channel
                tokens = l.split()[6:]
                df = pd.read_table(f, delim_whitespace=True, index_col=1, comment='%', names=['name','n','parity','flavor','weight'], usecols=range(1,6))
                # drop coherent scattering channels, which have different binning
                df = df[df["name"].str.startswith('coh_') == False]
            else:
                # Format 3: no binning specified, use SNOwGLoBES default values
                tokens = '% 200 0.0005 0.100 200 0.0005 0.100'.split()[1:]
                df = pd.read_table(f, delim_whitespace=True, index_col=1, comment='%', names=['name','n','parity','flavor','weight'])

            self.channels[material] = df

            nsamples, smin, smax, nbins, emin, emax = [float(t) for t in tokens]
            self.binning[material] = {'e_true': np.linspace(smin, smax, int(nsamples)+1),
                                      'e_smear': np.linspace(emin, emax, int(nbins)+1)
                                      }
        self.materials = list(self.channels.keys())
        self.chan_dir = chan_dir
        logger.info(f'read channels for materials: {self.materials}')
        logger.debug(f'channels: {self.channels}')

    def _load_efficiency_vectors(self, path):
        result = {}
        for detector in self.detectors:
            res_det = {}
            for file in (path/str(detector)).glob(f'effic_*_{detector}.dat'):
                channel =  file.stem[len('effic_'):-len(detector)-1]
                logger.debug(f'Reading file ({detector},{channel}): {file}')
                with open(file) as f:
                    effs = np.fromiter(f.readlines()[0].split("{")[-1].split("}")[0].split(","), float)
                    res_det[channel]= effs
            result[detector]=res_det
        self.efficiencies = result 
        logger.info(f'read efficiencies for detectors: {list(self.efficiencies.keys())}')
        logger.debug(f'efficiencies: {self.efficiencies}')

    def _load_smearing_matrices(self, path):
        result = {}
        for detector in self.detectors:
            res_det = {}
            for file in (path/str(detector)).glob(f'smear*_{detector}.dat'):
                channel =  file.stem[len('smear_'):-len(detector)-1]
                with open(file) as f:
                    lines = f.readlines()[1:-1]
                    while not "{" in lines[0]: lines = lines[1:]
                    while not "{" in lines[-1]: lines = lines[:-1]
                    matrix = np.zeros((len(lines),len(lines)))
                    for i,l in enumerate(lines):
                        elements = np.fromiter(l.split("{")[-1].split("}")[0].split(","), float)
                        matrix[i, int(elements[0]+0.1):int(elements[1]+0.1)+1] = elements[2:]
                    res_det[channel]= matrix
            result[detector]=res_det
        self.smearings = result 
        logger.info(f'read smearing matrices for detectors: {list(self.smearings.keys())}')
        logger.debug(f'smearing matrices: {self.smearings}')

    def compute_rates(self, detector:str, material:str, fluxes:np.ndarray, flux_energies:np.ndarray):
        """ Calculate the rates for the given neutrino fluxes interacting in the given detector.

        Parameters
        ----------
        detector: str
            Detector name in SNOwGLoBES. Check `SimpleRate.detectors` for the list of options
        material: str
            Material name in SNOwGLoBES. Check `SimpleRate.materials` for the list of options
        fluxes:
            2d array of the neutrino fluxes.
            First array dimension corresponds to flavors [nu_e, nu_mu, anti_nu_e, anti_nu_mu]
            Second array dimension corresponds to the `energies` bins,

        flux_energies:
            1d array of the neutrino energy bins in GeV, corresponding to the `fluxes`

        Returns
        -------
        pd.DataFrame
            Table with Energy (GeV) as index values,
            and number of events for each energy bin, for all interaction channels.
            Columns are hierarchical: (is_weighted, channel),
            so one can easily access the desired final table. 
        """
        TargetMass = self.detectors[detector].tgt_mass
        data = {}

        #load the binning for the smearing
        binning= self.binning[material]
        #calculate bin centers 
        energies_t = 0.5*(binning['e_true'][1:]+ binning['e_true'][:-1] )
        energies_s= 0.5*(binning['e_smear'][1:]+binning['e_smear'][:-1])
        binsize = np.diff(binning['e_true'])

        for channel in self.channels[material].itertuples():
            xsec_path = f"xscns/xs_{channel.name}.dat"
            xsec = np.loadtxt(self.base_dir/xsec_path)
            flavor_index = 0 if 'e' in channel.flavor else (1 if 'm' in channel.flavor else 2)
            flavor = flavor_index + (3 if channel.parity == '-' else 0)
            flux = fluxes[flavor]
            # Cross-section in 10^-38 cm^2
            xsecs = np.interp(np.log(energies_t)/np.log(10), xsec[:, 0], xsec[:, 1+flavor], left=0, right=0) * energies_t
            # Fluence (flux integrated over time bin) in cm^-2 
            # (must be divided by 0.2 MeV to compensate the multiplication in generate_time_series)
            fluxs = np.interp(energies_t, flux_energies, flux, left=0, right=0)/2e-4
            # Rate computation
            rates = xsecs * 1e-38 * fluxs * float(TargetMass) * 1./1.661e-33 * binsize
            # Weighting
            weighted_rates = rates * channel.weight
            # Write to dictionary
            data[(channel.name,'unsmeared','unweighted')] = rates
            data[(channel.name,'unsmeared','weighted')] = weighted_rates
            # Add detector effects
            if self.smearings and self.efficiencies:
                smear = self.smearings[detector].get(channel.name, np.eye(len(rates)))
                effic = self.efficiencies[detector].get(channel.name, np.ones(len(rates)))
                rates = np.dot(smear,rates) * effic
                weighted_rates = rates * channel.weight
                # Write to dictionary
                data[(channel.name,'smeared','unweighted')] = rates
                data[(channel.name,'smeared','weighted')] = weighted_rates
        #collect everything to pandas DataFrame
        df = pd.DataFrame(data, index = energies_s)
        df.index.rename('E', inplace=True)
        df.columns.rename(['channel','is_smeared','is_weighted'], inplace=True)
        return df.reorder_levels([2,1,0], axis='columns')
       
    def run(self, flux_files, detector:str, material:str=None):
        """ Compute expected rates for given configuration,
        collect the resulting data and return it in `pandas.DataFrame`

        Parameters
        -----------
        flux_files: list(str) or str
            An iterable of flux table filenames to process, or a single filename
        detector: str
            Detector name, SNOwGLoBES style
        material: str or None
            Material name, SNOwGLoBES style. If None, we'll try to guess it

        Returns
        --------
        list(pd.DataFrame or Exception)
            List with the data table for each flux_file, keeping the order.
            Each table containing Energy (GeV) as index values, 
            and number of events for each energy bin, for all interaction channels.
            Columns are hierarchical: (is_weighted, channel),
            so one can easily access the desired final table. 
            If run failed with exception, this exception will be returned (not raised).

        Raises
        ------
        ValueError
            if material or detector value is invalid

        """
        if not detector in self.detectors:
            raise ValueError(f'Detector "{detector}" is not in {list(self.detectors)}')
        if material is None:
            material = guess_material(detector)
        if not material in self.materials:
            raise ValueError(f'Material "{material}" is not in {self.materials}')
        if isinstance(flux_files,str):
            flux_files = [flux_files]

        with tqdm(total=len(flux_files), leave=False, desc='Flux files') as progressbar:
            results = []
            for flux_file in flux_files:
                e_flux = np.loadtxt(Path(flux_file).resolve()).T
                energies = e_flux[0]
                fluxes = e_flux[1:]
                result = self.compute_rates(detector, material, fluxes, energies)
                progressbar.update()
                results.append(result)
            return results



import unittest

def snewpy_test_suite():
    """Returns unittest.TestSuite of desiutil tests.

    This is factored out separately from runtests() so that it can be used by
    ``python setup.py test``.
    """
    from os.path import dirname
    pydir = dirname(dirname(__file__))
    tests = unittest.defaultTestLoader.discover(pydir,
                                                top_level_dir=dirname(pydir))
    return tests

def runtests():
    """Run all tests in snewpy.test.test_*.
    """
    # Load all TestCase classes from snewpy/test/test_*.py
    tests = snewpy_test_suite()
    # Run them
    unittest.TextTestRunner(verbosity=2).run(tests)

if __name__ == "__main__":
    runtests()



import itertools as it
import os
from abc import ABC, abstractmethod
from warnings import warn

import numpy as np
from astropy import units as u
from astropy.table import Table, join
from astropy.units import UnitTypeError, get_physical_type
from astropy.units.quantity import Quantity
from scipy.special import loggamma
from snewpy import _model_downloader

from snewpy.neutrino import Flavor
from snewpy.flavor_transformation import NoTransformation
from functools import wraps


def _wrap_init(init, check):
    @wraps(init)
    def _wrapper(self, *arg, **kwargs):
        init(self, *arg, **kwargs)
        check(self)
    return _wrapper


class SupernovaModel(ABC):
    """Base class defining an interface to a supernova model."""

    def __init_subclass__(cls, **kwargs):
        """Hook to modify the subclasses on creation"""
        super().__init_subclass__(**kwargs)
        cls.__init__ = _wrap_init(cls.__init__, cls.__post_init_check)

    def __init__(self, time, metadata):
        """Initialize supernova model base class
        (call this method in the subclass constructor as ``super().__init__(time,metadata)``).

        Parameters
        ----------
        time : ndarray of astropy.Quantity
            Time points where the model flux is defined.
            Must be array of :class:`Quantity`, with units convertable to "second".
        metadata : dict
            Dict of model parameters <name>:<value>,
            to be used for printing table in :meth:`__repr__` and :meth:`_repr_markdown_`
        """
        self.time = time
        self.metadata = metadata
        
    def __repr__(self):
        """Default representation of the model.
        """

        mod = f"{self.__class__.__name__} Model"
        try:
            mod += f': {self.filename}'
        except AttributeError:
            pass
        s = [mod]
        for name, v in self.metadata.items():
            s += [f"{name:16} : {v}"]
        return '\n'.join(s)

    def __post_init_check(self):
        """A function to check model integrity after initialization"""
        try:
            t = self.time
            m = self.metadata
        except AttributeError as e:
            clsname = self.__class__.__name__
            raise TypeError(f"Model not initialized. Please call 'SupernovaModel.__init__' within the '{clsname}.__init__'") from e

    def _repr_markdown_(self):
        """Markdown representation of the model, for Jupyter notebooks.
        """
        mod = f'**{self.__class__.__name__} Model**'
        try:
            mod +=f': {self.filename}'
        except:
            pass
        s = [mod,'']
        if self.metadata:
            s += ['|Parameter|Value|',
                  '|:--------|:----:|']
            for name, v in self.metadata.items():
                try: 
                    s += [f"|{name} | ${v.value:g}$ {v.unit:latex}|"]
                except:
                    s += [f"|{name} | {v} |"]
        return '\n'.join(s)

    def get_time(self):
        """Returns
        -------
        ndarray of astropy.Quantity
            Snapshot times from the simulation
        """
        return self.time

    @abstractmethod
    def get_initial_spectra(self, t, E, flavors=Flavor):
        """Get neutrino spectra at the source.

        Parameters
        ----------
        t : astropy.Quantity
            Time to evaluate initial spectra.
        E : astropy.Quantity or ndarray of astropy.Quantity
            Energies to evaluate the initial spectra.
        flavors: iterable of snewpy.neutrino.Flavor
            Return spectra for these flavors only (default: all)

        Returns
        -------
        initialspectra : dict
            Dictionary of neutrino spectra, keyed by neutrino flavor.
        """
        pass

    def get_initialspectra(self, *args):
        """DO NOT USE! Only for backward compatibility!

        :meta private:
        """
        warn("Please use `get_initial_spectra()` instead of `get_initialspectra()`!", FutureWarning)
        return self.get_initial_spectra(*args)

    def get_transformed_spectra(self, t, E, flavor_xform):
        """Get neutrino spectra after applying oscillation.

        Parameters
        ----------
        t : astropy.Quantity
            Time to evaluate initial and oscillated spectra.
        E : astropy.Quantity or ndarray of astropy.Quantity
            Energies to evaluate the initial and oscillated spectra.
        flavor_xform : FlavorTransformation
            An instance from the flavor_transformation module.

        Returns
        -------
        dict
            Dictionary of transformed spectra, keyed by neutrino flavor.
        """
        initialspectra = self.get_initial_spectra(t, E)
        transformed_spectra = {}

        transformed_spectra[Flavor.NU_E] = \
            flavor_xform.prob_ee(t, E) * initialspectra[Flavor.NU_E] + \
            flavor_xform.prob_ex(t, E) * initialspectra[Flavor.NU_X]

        transformed_spectra[Flavor.NU_X] = \
            flavor_xform.prob_xe(t, E) * initialspectra[Flavor.NU_E] + \
            flavor_xform.prob_xx(t, E) * initialspectra[Flavor.NU_X] 

        transformed_spectra[Flavor.NU_E_BAR] = \
            flavor_xform.prob_eebar(t, E) * initialspectra[Flavor.NU_E_BAR] + \
            flavor_xform.prob_exbar(t, E) * initialspectra[Flavor.NU_X_BAR]

        transformed_spectra[Flavor.NU_X_BAR] = \
            flavor_xform.prob_xebar(t, E) * initialspectra[Flavor.NU_E_BAR] + \
            flavor_xform.prob_xxbar(t, E) * initialspectra[Flavor.NU_X_BAR] 

        return transformed_spectra   

    def get_flux (self, t, E, distance, flavor_xform=NoTransformation()):
        """Get neutrino flux through 1cm^2 surface at the given distance

        Parameters
        ----------
        t : astropy.Quantity
            Time to evaluate the neutrino spectra.
        E : astropy.Quantity or ndarray of astropy.Quantity
            Energies to evaluate the the neutrino spectra.
        distance : astropy.Quantity or float (in kpc)
            Distance from supernova.
        flavor_xform : FlavorTransformation
            An instance from the flavor_transformation module.

        Returns
        -------
        dict
            Dictionary of neutrino fluxes in [neutrinos/(cm^2*erg*s)], 
            keyed by neutrino flavor.

        """
        distance = distance << u.kpc #assume that provided distance is in kpc, or convert
        factor = 1/(4*np.pi*(distance.to('cm'))**2)
        flux = self.get_transformed_spectra(t, E, flavor_xform)
        return {flavor: f*factor for flavor,f in flux.items()}



    def get_oscillatedspectra(self, *args):
        """DO NOT USE! Only for backward compatibility!

        :meta private:
        """
        warn("Please use `get_transformed_spectra()` instead of `get_oscillatedspectra()`!", FutureWarning)
        return self.get_transformed_spectra(*args)

def get_value(x):
    """If quantity x has is an astropy Quantity with units, return just the
    value.

    Parameters
    ----------
    x : Quantity, float, or ndarray
        Input quantity.

    Returns
    -------
    value : float or ndarray
    
    :meta private:
    """
    if type(x) == Quantity:
        return x.value
    return x

class PinchedModel(SupernovaModel):
    """Subclass that contains spectra/luminosity pinches"""
    def __init__(self, simtab, metadata):
        """ Initialize the PinchedModel using the data from the given table.

        Parameters
        ----------
        simtab: astropy.Table 
            Should contain columns TIME, {L,E,ALPHA}_NU_{E,E_BAR,X,X_BAR}
            The values for X_BAR may be missing, then NU_X data will be used
        metadata: dict
            Model parameters dict
        """
        if not 'L_NU_X_BAR' in simtab.colnames:
            # table only contains NU_E, NU_E_BAR, and NU_X, so double up
            # the use of NU_X for NU_X_BAR.
            for val in ['L','E','ALPHA']:
                simtab[f'{val}_NU_X_BAR'] = simtab[f'{val}_NU_X']
        # Get grid of model times.
        time = simtab['TIME'] << u.s
        # Set up dictionary of luminosity, mean energy and shape parameter
        # alpha, keyed by neutrino flavor (NU_E, NU_X, NU_E_BAR, NU_X_BAR).
        self.luminosity = {}
        self.meanE = {}
        self.pinch = {}
        for f in Flavor:
            self.luminosity[f] = simtab[f'L_{f.name}'] << u.erg/u.s
            self.meanE[f] = simtab[f'E_{f.name}'] << u.MeV
            self.pinch[f] = simtab[f'ALPHA_{f.name}']
        super().__init__(time, metadata)


    def get_initial_spectra(self, t, E, flavors=Flavor):
        """Get neutrino spectra/luminosity curves before oscillation.

        Parameters
        ----------
        t : astropy.Quantity
            Time to evaluate initial spectra.
        E : astropy.Quantity or ndarray of astropy.Quantity
            Energies to evaluate the initial spectra.
        flavors: iterable of snewpy.neutrino.Flavor
            Return spectra for these flavors only (default: all)

        Returns
        -------
        initialspectra : dict
            Dictionary of model spectra, keyed by neutrino flavor.
        """
        #convert input arguments to 1D arrays
        t = u.Quantity(t, ndmin=1)
        E = u.Quantity(E, ndmin=1)
        #Reshape the Energy array to shape [1,len(E)]
        E = np.expand_dims(E, axis=0)

        initialspectra = {}

        # Avoid division by zero in energy PDF below.
        E[E==0] = np.finfo(float).eps * E.unit

        # Estimate L(t), <E_nu(t)> and alpha(t). Express all energies in erg.
        E = E.to_value('erg')

        # Make sure input time uses the same units as the model time grid, or
        # the interpolation will not work correctly.
        t = t.to(self.time.unit)

        for flavor in flavors:
            # Use np.interp rather than scipy.interpolate.interp1d because it
            # can handle dimensional units (astropy.Quantity).
            L  = get_value(np.interp(t, self.time, self.luminosity[flavor].to('erg/s')))
            Ea = get_value(np.interp(t, self.time, self.meanE[flavor].to('erg')))
            a  = np.interp(t, self.time, self.pinch[flavor])

            #Reshape the time-related arrays to shape [len(t),1]
            L  = np.expand_dims(L, axis=1)
            Ea = np.expand_dims(Ea,axis=1)
            a  = np.expand_dims(a, axis=1)
            # For numerical stability, evaluate log PDF and then exponentiate.
            result = \
              np.exp(np.log(L) - (2+a)*np.log(Ea) + (1+a)*np.log(1+a)
                    - loggamma(1+a) + a*np.log(E) - (1+a)*(E/Ea)) / (u.erg * u.s)
            #remove bad values
            result[np.isnan(result)] = 0
            #remove unnecessary dimensions, if E or t was scalar:
            result = np.squeeze(result)
            initialspectra[flavor] = result
        return initialspectra


class _GarchingArchiveModel(PinchedModel):
    """Subclass that reads models in the format used in the
    `Garching Supernova Archive <https://wwwmpa.mpa-garching.mpg.de/ccsnarchive/>`_."""
    def __init__(self, filename, eos='LS220', metadata={}):
        """Model Initialization.

        Parameters
        ----------
        filename : str
            Absolute or relative path to file with model data, we add nue/nuebar/nux.  This argument will be deprecated.
        eos: str
            Equation of state. Valid value is 'LS220'. This argument will be deprecated.

        Other Parameters
        ----------------
        progenitor_mass: astropy.units.Quantity
            Mass of model progenitor in units Msun. Valid values are {progenitor_mass}.
        Raises
        ------
        FileNotFoundError
            If a file for the chosen model parameters cannot be found
        ValueError
            If a combination of parameters is invalid when loading from parameters
        """
        if not metadata:
            metadata = {
                'Progenitor mass': float(os.path.basename(filename).split('s')[1].split('c')[0]) * u.Msun,
                'EOS': eos,
            }

        # Read through the several ASCII files for the chosen simulation and
        # merge the data into one giant table.
        mergtab = None
        for flavor in Flavor:
            _flav = Flavor.NU_X if flavor == Flavor.NU_X_BAR else flavor
            _sfx = _flav.name.replace('_', '').lower()
            _filename = '{}_{}_{}'.format(filename, eos, _sfx)
            _lname = 'L_{}'.format(flavor.name)
            _ename = 'E_{}'.format(flavor.name)
            _e2name = 'E2_{}'.format(flavor.name)
            _aname = 'ALPHA_{}'.format(flavor.name)

            # Open the requested filename using the model downloader.
            datafile = _model_downloader.get_model_data(self.__class__.__name__, _filename)

            simtab = Table.read(datafile,
                                names=['TIME', _lname, _ename, _e2name],
                                format='ascii')
            simtab['TIME'].unit = 's'
            simtab[_lname].unit = '1e51 erg/s'
            simtab[_aname] = (2*simtab[_ename]**2 - simtab[_e2name]) / (simtab[_e2name] - simtab[_ename]**2)
            simtab[_ename].unit = 'MeV'
            del simtab[_e2name]

            if mergtab is None:
                mergtab = simtab
            else:
                mergtab = join(mergtab, simtab, keys='TIME', join_type='left')
                mergtab[_lname].fill_value = 0.
                mergtab[_ename].fill_value = 0.
                mergtab[_aname].fill_value = 0.
        simtab = mergtab.filled()
        super().__init__(simtab, metadata)


class _RegistryModel(ABC):
    """Base class for supernova model classes that initialise from physics parameters."""

    _param_validator = None

    @classmethod
    def get_param_combinations(cls):
        """Returns all valid combinations of parameters for a given SNEWPY register model.

        Subclasses can provide a Callable `cls._param_validator` that takes a combination of parameters
        as an argument and returns True if a particular combinations of parameters is valid.
        If None is provided, all combinations are considered valid.

        Returns
        -------
        valid_combinations: tuple[dict]
            A tuple of all valid parameter combinations stored as Dictionaries
        """
        for key, val in cls.param.items():
            if not isinstance(val, (list, Quantity)):
                cls.param[key] = [val]
            elif isinstance(val, Quantity) and val.size == 1:
                try:
                    # check if val.value is iterable, e.g. a list or a NumPy array
                    iter(val.value)
                except:
                    cls.param[key] = [val.value] * val.unit
        combos = tuple(dict(zip(cls.param, combo)) for combo in it.product(*cls.param.values()))
        return tuple(c for c in filter(cls._param_validator, combos))

    def check_valid_params(cls, **user_params):
        """Checks that the model-specific values, units, names and conbinations of requested parameters are valid.

        Parameters
        ----------
        user_params : varies
            User-requested model parameters to be tested for validity.
            NOTE: This must be provided as kwargs that match the keys of cls.param

        Raises
        ------
        ValueError
            If invalid model parameters are provided based on units, allowed values, etc.
        UnitTypeError
            If invalid units are provided for a model parameter

        See Also
        --------
        snewpy.models.ccsn
        snewpy.models.presn
        """
        # Check that the appropriate number of params are provided
        if not all(key in user_params for key in cls.param.keys()):
            raise ValueError(f"Missing parameter! Expected {cls.param.keys()} but was given {user_params.keys()}")

        # Check parameter units and values
        for (key, allowed_params), user_param in zip(cls.param.items(), user_params.values()):

            # If both have units, check that the user param value is valid. If valid, continue. Else, error
            if type(user_param) == Quantity and type(allowed_params) == Quantity:
                if get_physical_type(user_param.unit) != get_physical_type(allowed_params.unit):
                    raise UnitTypeError(f"Incorrect units {user_param.unit} provided for parameter {key}, "
                                        f"expected {allowed_params.unit}")

                elif np.isin(user_param.to(allowed_params.unit).value, allowed_params.value):
                    continue
                else:
                    raise ValueError(f"Invalid value '{user_param}' provided for parameter {key}, "
                                     f"allowed value(s): {allowed_params}")

            # If one only one has units, then error
            elif (type(user_param) == Quantity) ^ (type(allowed_params) == Quantity):
                # User param has units, model param is unitless
                if type(user_param) == Quantity:
                    raise ValueError(f"Invalid units {user_param.unit} for parameter {key} provided, expected None")
                else:
                    raise ValueError(f"Missing units for parameter {key}, expected {allowed_params.unit}")

            # Check that unitless user param value is valid. If valid, continue. Else, Error
            elif user_param in allowed_params:
                continue
            else:
                raise ValueError(f"Invalid value '{user_param}' provided for parameter {key}, "
                                 f"allowed value(s): {allowed_params}")

        # Check Combinations (Logic lives inside model subclasses under model.isvalid_param_combo)
        if user_params not in cls.get_param_combinations():
            raise ValueError(
                f"Invalid parameter combination. See {cls.__class__.__name__}.get_param_combinations() for a "
                "list of allowed parameter combinations.")



# -*- coding: utf-8 -*-
"""Utility module that reads a local YAML file listing the data files
available for the models supported by SNEWPY. These data files can either
be local (in the SNEWPY source tree) or uploaded to a model repository on
Zenodo. The YAML file supports regular expressions to allow matching of all
possible model files.
"""

import hashlib
import os
import requests
import yaml

from dataclasses import dataclass
from importlib.resources import open_text
from pathlib import Path
from tqdm.auto import tqdm
from typing import Optional

from snewpy import model_path

import logging
logger = logging.getLogger('FileHandle')


def _md5(fname:str) -> str:
    """calculate the md5sum hash of a file."""
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def _download(src:str, dest:str, chunk_size=8192):
    """Download a file from 'src' to 'dest' and show the progress bar."""
    #make sure parent dir exists
    Path(dest).parent.mkdir(exist_ok=True, parents=True)
    with requests.get(src, stream=True) as r:
        r.raise_for_status()
        fileSize = int(r.headers.get('content-length', 0))
        with tqdm(desc=dest.name, total=fileSize, unit='iB', unit_scale=True, unit_divisor=1024) as bar:
            with open(dest, 'wb') as f:
                for chunk in r.iter_content(chunk_size=chunk_size):
                    size = f.write(chunk)
                    bar.update(size)


class ChecksumError(FileNotFoundError):
    """Raise an exception due to a mismatch in the MD5 checksum.
    """
    def __init__(self, path, md5_exp, md5_actual):
        super().__init__(f'Checksum error for file {path}: {md5_actual}!={md5_exp}')
    pass


class MissingFileError(FileNotFoundError):
    """Raise an exception due to a missing file.
    """
    pass


@dataclass
class FileHandle:
    """Object storing local path, remote URL (optional), and MD5 sum
(optional) for a SNEWPY model file. If the requested file is already present
locally, open it. Otherwise, download it from the remote URL to the desired
local path.
    """
    path: Path
    remote: str = None
    md5: Optional[str] = None
    
    def check(self) -> None:
        """Check if the given file exists locally and has a correct md5 sum.
        Raises
        ------
        :class:`MissingFileError`
            if the local copy of the file is missing
        :class:`ChecksumError` 
            if the local file exists, but the checksum is wrong"""
        if not self.path.exists():
            raise MissingFileError(self.path)
        if self.md5:
            logger.info(f'File {self.path}: checking md5')
            md5 = _md5(self.path)
            logger.debug(f'{md5} vs expected {self.md5}')
            if (md5 != self.md5):
                raise ChecksumError(self.path, self.md5, md5)
    
    def load(self) -> Path:
        """Make sure that local file exists and has a correct checksum.
        Download the file if needed.
        """
        try:
            self.check()
        except FileNotFoundError as e:
            logger.info(f'Downloading file {self.path}')
            _download(self.remote, self.path)
            self.check()
        return self.path


def from_zenodo(zenodo_id:str, model:str, filename:str, path:str=model_path):
    """Access files on Zenodo.

    Parameters
    ----------
    zenodo_id : Zenodo record for model files.
    model : Name of the model class for this model file.
    filename : Expected filename storing simulation data.
    path : Local installation path (defaults to astropy cache).

    Returns
    -------
    file : FileHandle object.
    """
    zenodo_url = f'https://zenodo.org/api/records/{zenodo_id}'
    path = Path(path)/str(model)
    path.mkdir(exist_ok=True, parents=True)

    record = requests.get(zenodo_url).json()

    # Search for model file string in Zenodo request for this record.
    file = next((_file for _file in record['files'] if _file['key'] == filename), None)

    # If matched, return a FileHandle which takes care of downloading and
    # checksum (or loads a local data file). Otherwise raise an exception.
    if file is not None:
        return FileHandle(path = path/file['key'],
                                 remote= file['links']['self'],
                                 md5 = file['checksum'].split(':')[1])
    else:
        raise MissingFileError(filename)


def from_github(release_version:str, model:str, filename:str, path:str=model_path):
    """Access files on GitHub.

    Parameters
    ----------
    release_version: SNEWPY release with corresponding model data.
    model : Name of the model class for this model file.
    filename : Expected filename storing simulation data.
    path : Local installation path (defaults to astropy cache).

    Returns
    -------
    file : FileHandle object.
    """
    github_url = f'https://github.com/SNEWS2/snewpy/raw/v{release_version}/models/{model}/{filename}'
    localpath = Path(path)/str(model)
    localpath.mkdir(exist_ok=True, parents=True)

    return FileHandle(path = localpath/filename,
                      remote = github_url)


def get_model_data(model: str, filename: str, path: str = model_path) -> Path:
    """Access model data. Configuration for each model is in a YAML file
    distributed with SNEWPY.

    Parameters
    ----------
    model : Name of the model class for this model file.
    filename : Name of simulation datafile, or a relative path from the model sub-directory
    path : Local installation path (defaults to astropy cache).

    Returns
    -------
    Path of downloaded file.
    """
    if os.path.isabs(filename):
        return Path(filename)

    params = { 'model':model, 'filename':filename, 'path':path }

    # Parse YAML file with model repository configurations.
    with open_text('snewpy.models', 'model_files.yml') as f:
        config = yaml.safe_load(f)
        models = config['models']
        # Search for model in YAML configuration.
        if model in models.keys():
            # Get data from GitHub or Zenodo.
            modconf = models[model]
            repo = modconf['repository']

            if repo == 'github':
                params['release_version'] = modconf['release_version']
                fh = from_github(**params)
            elif repo == 'zenodo':
                params['zenodo_id'] = modconf['zenodo_id']
                fh = from_zenodo(**params)
            else:
                raise ValueError(f'Repository {repo} not recognized')
            return fh.load()
        else:
            raise KeyError(f'No configuration for {model}')




# -*- coding: utf-8 -*-
"""The ``snewpy.snowglobes`` module contains functions for interacting with SNOwGLoBES.

`SNOwGLoBES <https://github.com/SNOwGLoBES/snowglobes>`_ can estimate detected
event rates from a given input supernova neutrino flux. It supports many
different neutrino detectors, detector materials and interaction channels.
There are three basic steps to using SNOwGLoBES from SNEWPY:

* **Generating input files for SNOwGLoBES:**
    There are two ways to do this, either generate a time series or a fluence file. This is done taking as input the supernova simulation model.
    The first will evaluate the neutrino flux at each time step, the latter will compute the integrated neutrino flux (fluence) in the time bin.
    The result is a compressed .tar file containing all individual input files.
* **Running SNOwGLoBES:**
    This step convolves the fluence generated in the previous step with the cross-sections for the interaction channels happening in various detectors supported by SNOwGLoBES.
    It takes into account the effective mass of the detector as well as a smearing matrix describing the energy-dependent detection efficiency.
    The output gives the number of events detected as a function of energy for each interaction channel, integrated in a given time window (or time bin), or in a snapshot in time.
* **Collating SNOwGLoBES outputs:**
    This step puts together all the interaction channels and time bins evaluated by SNOwGLoBES in a single file (for each detector and for each time bin).
    The output tables allow to build the detected neutrino energy spectrum and neutrino time distribution, for each reaction channel or the sum of them.
"""

import io
import logging
import os
import re
import tarfile
from pathlib import Path
from tempfile import TemporaryDirectory

import matplotlib.pyplot as plt
import numpy as np
from astropy import units as u
from tqdm.auto import tqdm
from warnings import warn

import snewpy.models
from snewpy.flavor_transformation import *
from snewpy.neutrino import Flavor, MassHierarchy
from snewpy.snowglobes_interface import SimpleRate

logger = logging.getLogger(__name__)

def generate_time_series(model_path, model_type, transformation_type, d, output_filename=None, ntbins=30, deltat=None, snmodel_dict={}):
    """Generate time series files in SNOwGLoBES format.

    This version will subsample the times in a supernova model, produce energy
    tables expected by SNOwGLoBES, and compress the output into a tarfile.

    Parameters
    ----------
    model_path : str
        Input file containing neutrino flux information from supernova model.
    model_type : str
        Format of input file. Matches the name of the corresponding class in :py:mod:`snewpy.models`.
    transformation_type : str
        Name of flavor transformation. See snewpy.flavor_transformation documentation for possible values.
    d : int or float
        Distance to supernova in kpc.
    output_filename : str or None
        Name of output file. If ``None``, will be based on input file name.
    ntbins : int
        Number of time slices. Will be ignored if ``deltat`` is also given.
    deltat : astropy.Quantity or None
        Length of time slices.
    snmodel_dict : dict
        Additional arguments for setting up the supernova model. See documentation of relevant ``SupernovaModel`` subclass for available options. (Optional)

    Returns
    -------
    str
        Path of compressed .tar file with neutrino flux data.
    """
    model_class = getattr(snewpy.models.ccsn, model_type)

    # Choose flavor transformation. Use dict to associate the transformation name with its class.
    flavor_transformation_dict = {'NoTransformation': NoTransformation(), 'AdiabaticMSW_NMO': AdiabaticMSW(mh=MassHierarchy.NORMAL), 'AdiabaticMSW_IMO': AdiabaticMSW(mh=MassHierarchy.INVERTED), 'NonAdiabaticMSWH_NMO': NonAdiabaticMSWH(mh=MassHierarchy.NORMAL), 'NonAdiabaticMSWH_IMO': NonAdiabaticMSWH(mh=MassHierarchy.INVERTED), 'TwoFlavorDecoherence': TwoFlavorDecoherence(), 'ThreeFlavorDecoherence': ThreeFlavorDecoherence(), 'NeutrinoDecay_NMO': NeutrinoDecay(mh=MassHierarchy.NORMAL), 'NeutrinoDecay_IMO': NeutrinoDecay(mh=MassHierarchy.INVERTED)}
    flavor_transformation = flavor_transformation_dict[transformation_type]

    model_dir, model_file = os.path.split(os.path.abspath(model_path))
    snmodel = model_class(model_path, **snmodel_dict)

    # Subsample the model time. Default to 30 time slices.
    tmin = snmodel.get_time()[0]
    tmax = snmodel.get_time()[-1]
    if deltat is not None:
        dt = deltat
        ntbins = int((tmax-tmin)/dt)
    else:
        dt = (tmax - tmin) / (ntbins+1)

    tedges = np.arange(tmin/u.s, tmax/u.s, dt/u.s)*u.s
    times = 0.5*(tedges[1:] + tedges[:-1])

    # Generate output.
    if output_filename is not None:
        tfname = output_filename + 'kpc.tar.bz2'
    else:
        model_file_root, _ = os.path.splitext(model_file)  # strip extension (if present)
        tfname = model_file_root + '.' + transformation_type + '.{:.3f},{:.3f},{:d}-{:.1f}'.format(tmin, tmax, ntbins, d) + 'kpc.tar.bz2'

    with tarfile.open(os.path.join(model_dir, tfname), 'w:bz2') as tf:
        #creates file in tar archive that gives information on parameters
        output = '\n'.join(map(str, transformation_type)).encode('ascii')
        tf.addfile(tarfile.TarInfo(name='parameterinfo'), io.BytesIO(output))

        MeV = 1.60218e-6 * u.erg
        energy = np.linspace(0, 100, 501) * MeV  # 1MeV

        # Loop over sampled times.
        for i, t in enumerate(times):
            osc_spectra = snmodel.get_transformed_spectra(t, energy, flavor_transformation)

            osc_fluence = {}
            table = []

            table.append('# TBinMid={:g}sec TBinWidth={:g}s EBinWidth=0.2MeV Fluence at Earth for this timebin in neutrinos per cm^2'.format(t, dt))
            table.append('# E(GeV)	NuE	NuMu	NuTau	aNuE	aNuMu	aNuTau')

            # Generate energy + number flux table.
            for j, E in enumerate(energy):
                for flavor in Flavor:
                    osc_fluence[flavor] = osc_spectra[flavor][j] * dt * 0.2 * MeV / (4.*np.pi*(d*1000*3.086e+18)**2)

                s = '{:17.8E}'.format(E/(1e3 * MeV))
                s = '{}{:17.8E}'.format(s, osc_fluence[Flavor.NU_E])
                s = '{}{:17.8E}'.format(s, osc_fluence[Flavor.NU_X])
                s = '{}{:17.8E}'.format(s, osc_fluence[Flavor.NU_X])
                s = '{}{:17.8E}'.format(s, osc_fluence[Flavor.NU_E_BAR])
                s = '{}{:17.8E}'.format(s, osc_fluence[Flavor.NU_X_BAR])
                s = '{}{:17.8E}'.format(s, osc_fluence[Flavor.NU_X_BAR])
                table.append(s)
                logging.debug(s)

            # Encode energy/flux table and output to file in tar archive.
            output = '\n'.join(table).encode('ascii')

            extension = ".dat"
            model_file_root, _ = os.path.splitext(model_file)
            filename = model_file_root + '.tbin{:01d}.'.format(i+1) + transformation_type + \
                '.{:.3f},{:.3f},{:01d}-{:.1f}kpc{}'.format(tmin/u.s, tmax/u.s, ntbins, d, extension)

            info = tarfile.TarInfo(name=filename)
            info.size = len(output)
            tf.addfile(info, io.BytesIO(output))

    return os.path.join(model_dir, tfname)


def generate_fluence(model_path, model_type, transformation_type, d, output_filename=None, tstart=None, tend=None, snmodel_dict={}):
    """Generate fluence files in SNOwGLoBES format.

    This version will subsample the times in a supernova model, produce energy
    tables expected by SNOwGLoBES, and compress the output into a tarfile.

    Parameters
    ----------
    model_path : str
        Input file containing neutrino flux information from supernova model.
    model_type : str
        Format of input file. Matches the name of the corresponding class in :py:mod:`snewpy.models`.
    transformation_type : str
        Name of flavor transformation. See snewpy.flavor_transformation documentation for possible values.
    d : int or float
        Distance to supernova in kpc.
    output_filename : str or None
        Name of output file. If ``None``, will be based on input file name.
    tstart : astropy.Quantity or None
        Start of time interval to integrate over, or list of start times of the time series bins.
    tend : astropy.Quantity or None
        End of time interval to integrate over, or list of end times of the time series bins.
    snmodel_dict : dict
        Additional arguments for setting up the supernova model. See documentation of relevant ``SupernovaModel`` subclass for available options. (Optional)

    Returns
    -------
    str
        Path of compressed .tar file with neutrino flux data.
    """
    model_class = getattr(snewpy.models.ccsn, model_type)

    # Choose flavor transformation. Use dict to associate the transformation name with its class.
    flavor_transformation_dict = {'NoTransformation': NoTransformation(), 'AdiabaticMSW_NMO': AdiabaticMSW(mh=MassHierarchy.NORMAL), 'AdiabaticMSW_IMO': AdiabaticMSW(mh=MassHierarchy.INVERTED), 'NonAdiabaticMSWH_NMO': NonAdiabaticMSWH(mh=MassHierarchy.NORMAL), 'NonAdiabaticMSWH_IMO': NonAdiabaticMSWH(mh=MassHierarchy.INVERTED), 'TwoFlavorDecoherence': TwoFlavorDecoherence(), 'ThreeFlavorDecoherence': ThreeFlavorDecoherence(), 'NeutrinoDecay_NMO': NeutrinoDecay(mh=MassHierarchy.NORMAL), 'NeutrinoDecay_IMO': NeutrinoDecay(mh=MassHierarchy.INVERTED)}
    flavor_transformation = flavor_transformation_dict[transformation_type]

    model_dir, model_file = os.path.split(os.path.abspath(model_path))
    snmodel = model_class(model_path, **snmodel_dict)

    #set the timings up
    #default if inputs are None: full time window of the model
    if tstart is None:
        tstart = snmodel.get_time()[0]
        tend = snmodel.get_time()[-1]

    try:
        if len(tstart/u.s) > 0:
            t0 = tstart[0]
            t1 = tend[-1]
            nbin = len(tstart/u.s)
    except:
        t0 = tstart
        t1 = tend
        nbin = 1

    times = 0.5*(tstart + tend)

    model_times = snmodel.get_time()
    model_tstart = model_times*1.0
    model_tend = model_times*1.0

    model_tstart[0] = model_times[0]
    for i in range(1, len(model_times)):
        model_tstart[i] = 0.5*(model_times[i]+model_times[i-1])
        model_tend[i-1] = model_tstart[i]
    model_tend[len(model_times)-1] = model_times[-1]

    if nbin > 1:
        starting_index = np.zeros(len(times), dtype=np.int64)
        ending_index = np.zeros(len(times), dtype=np.int64)
        for i in range(len(tstart)):
            starting_index[i] = next(j for j, t in enumerate(model_tend) if t > tstart[i])
            ending_index[i] = next(j for j, t in enumerate(model_tend) if t >= tend[i])
    else:
        starting_index = [next(j for j, t in enumerate(model_tend) if t > tstart)]
        ending_index = [next(j for j, t in enumerate(model_tend) if t >= tend)]

    # Generate output.
    if output_filename is not None:
        tfname = output_filename+'.tar.bz2'
    else:
        model_file_root, _ = os.path.splitext(model_file)  # strip extension (if present)
        tfname = model_file_root + '.' + transformation_type + '.{:.3f},{:.3f},{:d}-{:.1f}'.format(t0, t1, nbin, d) + 'kpc.tar.bz2'

    with tarfile.open(os.path.join(model_dir, tfname), 'w:bz2') as tf:
        #creates file in tar archive that gives information on parameters
        output = '\n'.join(map(str, transformation_type)).encode('ascii')
        tf.addfile(tarfile.TarInfo(name='parameterinfo'), io.BytesIO(output))

        MeV = 1.60218e-6 * u.erg
        energy = np.linspace(0, 100, 501) * MeV

        # Loop over sampled times.
        for i in range(nbin):

            if nbin > 1:
                ta = tstart[i]
                tb = tend[i]
                t = times[i]
                dt = tb-ta
            else:
                ta = tstart
                tb = tend
                t = times
                dt = tb-ta

            #first time bin of model in requested interval
            osc_spectra = snmodel.get_transformed_spectra(model_times[starting_index[i]], energy, flavor_transformation)

            if dt >= model_tend[starting_index[i]]-ta:
                for flavor in Flavor:
                    osc_spectra[flavor] *= (model_tend[starting_index[i]]-ta)

                #intermediate time bins of model in requested interval
                for j in range(starting_index[i]+1, ending_index[i]):
                    temp_spectra = snmodel.get_transformed_spectra(model_times[j], energy, flavor_transformation)
                    for flavor in Flavor:
                        osc_spectra[flavor] += temp_spectra[flavor]*(model_tend[j]-model_tstart[j])

                #last time bin of model in requested interval
                temp_spectra = snmodel.get_transformed_spectra(
                    model_times[ending_index[i]], energy, flavor_transformation)
                for flavor in Flavor:
                    osc_spectra[flavor] += temp_spectra[flavor]*(tb-model_tstart[ending_index[i]])

                for flavor in Flavor:
                    osc_spectra[flavor] /= (tb-ta)

            osc_fluence = {}
            table = []

            table.append('# TBinMid={:g}sec TBinWidth={:g}s EBinWidth=0.2MeV Fluence at Earth for this timebin in neutrinos per cm^2'.format(t, dt))
            table.append('# E(GeV)	NuE	NuMu	NuTau	aNuE	aNuMu	aNuTau')

            # Generate energy + number flux table.
            for j, E in enumerate(energy):
                for flavor in Flavor:
                    osc_fluence[flavor] = osc_spectra[flavor][j] * dt * 0.2 * MeV / (4.*np.pi*(d*1000*3.086e+18)**2)

                s = '{:17.8E}'.format(E/(1e3 * MeV))
                s = '{}{:17.8E}'.format(s, osc_fluence[Flavor.NU_E])
                s = '{}{:17.8E}'.format(s, osc_fluence[Flavor.NU_X])
                s = '{}{:17.8E}'.format(s, osc_fluence[Flavor.NU_X])
                s = '{}{:17.8E}'.format(s, osc_fluence[Flavor.NU_E_BAR])
                s = '{}{:17.8E}'.format(s, osc_fluence[Flavor.NU_X_BAR])
                s = '{}{:17.8E}'.format(s, osc_fluence[Flavor.NU_X_BAR])
                table.append(s)
                logging.debug(s)

            # Encode energy/flux table and output to file in tar archive.
            output = '\n'.join(table).encode('ascii')

            extension = ".dat"
            if output_filename is not None:
                if nbin > 1:
                    filename = output_filename+"_"+str(i)+extension
                else:
                    filename = output_filename+extension
            else:
                model_file_root, _ = os.path.splitext(model_file)  # strip extension (if present)
                filename = model_file_root + '.tbin{:01d}.'.format(i+1) + transformation_type + \
                    '.{:.3f},{:.3f},{:01d}-{:.1f}kpc{}'.format(t0, t1, nbin, d, extension)

            info = tarfile.TarInfo(name=filename)
            info.size = len(output)
            tf.addfile(info, io.BytesIO(output))

    return os.path.join(model_dir, tfname)

def simulate(SNOwGLoBESdir, tarball_path, detector_input="all", verbose=False, *, detector_effects=True):
    """Calculate expected event rates for the given neutrino flux files and the given (set of) SNOwGLoBES detector(s).
    These event rates are given as a function of the neutrino energy and time, for each interaction channel.

    Parameters
    ----------
    SNOwGLoBESdir : str or None
        Path to SNOwGLoBES directory. Set to ``None`` to automatically use the latest supported SNOwGLoBES release.
    tarball_path : str
        Path of compressed .tar file produced e.g. by ``generate_time_series()`` or ``generate_fluence()``.
    detector_input : str
        Name of detector. If ``"all"``, will use all detectors supported by SNOwGLoBES.
    verbose : bool
        [DEPRECATED, DO NOT USE.]
    detector_effects : bool
         Whether to account for detector smearing and efficiency.
    """
    if verbose:  # Deprecated since SNEWPY v1.2
        warn(f"The 'verbose' parameter to 'snewpy.snowglobes.simulate()' is deprecated and should not be used.", FutureWarning)

    sng = SimpleRate(base_dir=SNOwGLoBESdir, detectors=detector_input, detector_effects=detector_effects)
    detector_input = list(sng.detectors)

    result = {}
    #Extracts data from tarfile and sets up lists of paths and fluxfilenames for later use
    with TemporaryDirectory(prefix='snowglobes') as tempdir:
        with tarfile.open(tarball_path) as tar:
            tar.extractall(tempdir)

        flux_files = list(Path(tempdir).glob('*.dat'))
        if len(detector_input)>0:
            detector_input = tqdm(detector_input, desc='Detectors', leave=False)
        for det in detector_input:
            res=sng.run(flux_files, det)
            result[det]=dict(zip((f.stem for f in flux_files),res))

    # save result to file for re-use in collate()
    cache_file = tarball_path[:tarball_path.rfind('.tar')] + '.npy'
    logging.info(f'Saving simulation results to {cache_file}')
    np.save(cache_file, result)
    return result 

re_chan_label = re.compile(r'nu(e|mu|tau)(bar|)_([A-Z][a-z]*)(\d*)_?(.*)')
def get_channel_label(c):
    mapp = {'nc':'NeutralCurrent',
            'ibd':'Inverse Beta Decay',
            'e':r'${\nu}_x+e^-$'}
    def gen_label(m):
        flv,bar,Nuc,num,res = m.groups()
        if flv!='e':
            flv='\\'+flv
        if bar:
            bar='\\'+bar
        s = f'${bar}{{\\nu}}_{flv}$ '+f'${{}}^{{{num}}}{Nuc}$ '+res
        return s

    if c in mapp:
        return mapp[c]
    else: 
        return re_chan_label.sub(gen_label, c) 

def collate(SNOwGLoBESdir, tarball_path, detector_input="", skip_plots=False, verbose=False, remove_generated_files=True, *, smearing=True):
    """Collates SNOwGLoBES output files and generates plots or returns a data table.

    Parameters
    ----------
    SNOwGLoBESdir : str or None
        [DEPRECATED, DO NOT USE.]
    tarball_path : str
        Path of compressed .tar file produced e.g. by ``generate_time_series()`` or ``generate_fluence()``.
    detector_input : str
        [DEPRECATED, DO NOT USE. SNEWPY will use all detectors included in the tarball.]
    skip_plots: bool
        If False, it gives as output the plot of the energy distribution for each time bin and for each interaction channel.
    verbose : bool
        [DEPRECATED, DO NOT USE.]
    remove_generated_files: bool
        [DEPRECATED, DO NOT USE.]
    smearing: bool
        Also consider results with smearing effects.

    Returns
    -------
    dict
        Dictionary of data tables: One table per time bin; each table contains in the first column the energy bins, in the remaining columns the number of events for each interaction channel in the detector.
    """
    if verbose:  # Deprecated since SNEWPY v1.2
        warn(f"The 'verbose' parameter to 'snewpy.snowglobes.collate()' is deprecated and should not be used.", FutureWarning)
    if detector_input:  # Deprecated since SNEWPY v1.2
        warn(f"The 'detector_input' parameter to 'snewpy.snowglobes.collate()' is deprecated and should not be used.", FutureWarning)
    if not remove_generated_files:  # Deprecated since SNEWPY v1.2
        warn(f"The 'remove_generated_files' parameter to 'snewpy.snowglobes.collate()' is deprecated and should not be used.", FutureWarning)

    def aggregate_channels(table, **patterns):
        #rearrange the table to have only channel column
        levels = list(table.columns.names)
        levels.remove('channel')
        t = table.stack(levels)
        for name,pattern in patterns.items():
            #get channels which contain `like`
            t_sel = t.filter(like=pattern)
            #sum over them and save to a separate column
            t_agg = t_sel.sum(axis='columns')
            #drop processed channels
            t.drop(t_sel.columns, axis='columns',inplace=True)
            t[name]=t_agg #fill the column
        #return table with the original levels order
        t = t.unstack(levels)
        t = t.reorder_levels(table.columns.names, axis=1)
        return t
        
    def do_plot(table, params):
        #plotting the events from given table
        flux,det,weighted,smeared = params
        for c in table.columns:
            if table[c].max() > 0.1:
                plt.plot(table[c],drawstyle='steps',label=get_channel_label(c), lw=1)
        plt.xlim(right=0.10)
        plt.ylim(bottom=0.10)
        plt.yscale('log')
        plt.legend(bbox_to_anchor=(0.5, 0.5, 0.5, 0.5), loc='best', borderaxespad=0)  # formats complete graph
        smear_title = 'Interaction' if smeared=='unsmeared' else 'Detected'
        plt.title(f'{flux} {det.capitalize()} {weighted.capitalize()} {smear_title} Events')
        if smeared=='smeared':
            plt.xlabel('Detected Energy (GeV)')
            plt.ylabel('Events')  
        else:
            plt.xlabel('Neutrino Energy (GeV)')
            plt.ylabel('Interaction Events')  

    #read the results from storage
    cache_file = tarball_path[:tarball_path.rfind('.tar')] + '.npy'
    logging.info(f'Reading tables from {cache_file}')
    tables = np.load(cache_file, allow_pickle=True).tolist()
    #This output is similar to what produced by:
    #tables = simulate(SNOwGLoBESdir, tarball_path,detector_input)

    #dict for old-style results, for backward compatibiity
    results = {}
    smearing_options = ['smeared','unsmeared'] if smearing else ['unsmeared']
    #save collated files:
    with TemporaryDirectory(prefix='snowglobes') as tempdir:
        tempdir = Path(tempdir)
        for det in tables:
            results[det] = {}
            for flux,t in tables[det].items():
                t = aggregate_channels(t,nc='nc_',e='_e')
                for w in ['weighted','unweighted']:
                    for s in smearing_options:
                        table = t[w][s]
                        filename_base = f'{flux}_{det}_events_{s}_{w}'
                        filename = tempdir/f'Collated_{filename_base}.dat'
                        #save results to text files
                        with open(filename,'w') as f:
                            f.write(table.to_string(float_format='%23.15g'))
                        #format the results for the output
                        header = 'Energy '+' '.join(list(table.columns))
                        data = table.to_numpy().T
                        index = table.index.to_numpy()
                        data = np.concatenate([[index],data])
                        results[filename.name] = {'header':header,'data':data}
                        #optionally plot the results
                        if skip_plots is False:
                            plt.figure(dpi=300)
                            do_plot(table,(flux,det,w,s))
                            filename = tempdir/f'{filename_base}_log_plot.png'
                            plt.savefig(filename, dpi=300, bbox_inches='tight')
                            plt.close()
        #Make a tarfile with the condensed data files and plots
        output_name = Path(tarball_path).stem
        output_name = output_name[:output_name.rfind('.tar')]+'_SNOprocessed'
        output_path = Path(tarball_path).parent/(output_name+'.tar.gz')
        with tarfile.open(output_path, "w:gz") as tar:
            for file in tempdir.iterdir():
                tar.add(file,arcname=output_name+'/'+file.name)
        logging.info(f'Created archive: {output_path}')
    return results 



# -*- coding: utf-8 -*-
"""
The submodule ``snewpy.models.ccsn`` contains models of core-collapse supernovae
derived from the :class:`SupernovaModel` base class.

Since SNEWPY v1.3, the prefered method is to initialise a model based on its physics parameters.
Initialisation from a file name is deprecated.

Use the ``param`` class property to view all physics parameters and their possible values:

>>> from snewpy.models.ccsn import Nakazato_2013
>>> Nakazato_2013.param
{'progenitor_mass': <Quantity [13., 20., 30., 50.] solMass>,
 'revival_time': <Quantity [  0., 100., 200., 300.] ms>,
 'metallicity': [0.02, 0.004],
 'eos': ['LS220', 'shen', 'togashi']}

For some models, not all combinations of parameters are valid. Use the ``get_param_combinations()``
class method to get a list of all valid combinations and filter it:

>>> list(params for params in Nakazato_2013.get_param_combinations() if params['eos'] != 'shen')
[{'progenitor_mass': <Quantity 30. solMass>, 'revival_time': <Quantity 0. ms>,
  'metallicity': 0.004, 'eos': 'LS220'},
 {'progenitor_mass': <Quantity 30. solMass>, 'revival_time': <Quantity 0. ms>,
  'metallicity': 0.004, 'eos': 'togashi'}]

.. _Garching Supernova Archive: https://wwwmpa.mpa-garching.mpg.de/ccsnarchive/
"""
import logging
import os
import re
import tarfile

import numpy as np
from astropy import units as u
from astropy.table import Table

from snewpy.models import loaders
from .base import PinchedModel, _RegistryModel

from warnings import warn
from functools import wraps


def _warn_deprecated_filename_argument(func):
    """Decorator for model.__new__ methods, causing them to issue a deprecation warning
     if argument `filename` is used. Initialization by filename will be moved to
     snewpy.models.init_model, and model classes are initialized from physical parameters.
    """
    @wraps(func)  # Ensures docstrings are preserved
    def decorator(cls, *args, **kwargs):
        filename = args[0] if len(args) > 0 else None  # Assumes filename is first pos. arg if provided
        if filename is not None:
            msg = ''.join(['Initializing this model with a filename is deprecated. ',
                           f'Instead, use keyword arguments {list(cls.param.keys())}. ',
                           f'See `{cls.__name__}.param`, `{cls.__name__}.get_param_combinations()` for more info.'])
            warn(FutureWarning(msg), stacklevel=2)
        return func(cls, *args, **kwargs)
    return decorator


class Analytic3Species(PinchedModel):
    """An analytical model calculating spectra given total luminosity,
    average energy, and rms or pinch, for each species.
    """

    param = "There are no input files available for this class. Use `doc/scripts/Analytic.py` in the SNEWPY GitHub repo to create a custom input file."

    def get_param_combinations(cls):
        print(cls.param)
        return []

    def __init__(self, filename):
        """
        Parameters
        ----------
        filename : str
            Absolute or relative path to file with model data.
        """

        simtab = Table.read(filename,format='ascii')
        self.filename = filename
        super().__init__(simtab, metadata={})


class Nakazato_2013(_RegistryModel):
    """Model based on simulations from Nakazato et al., ApJ S 205:2
    (2013), ApJ 804:75 (2015), PASJ 73:639 (2021). See also http://asphwww.ph.noda.tus.ac.jp/snn/.
    """

    param = {'progenitor_mass': [13, 20, 30, 50] * u.Msun,
             'revival_time': [0, 100, 200, 300] * u.ms,
             'metallicity': [0.02, 0.004],
             'eos': ['LS220', 'shen', 'togashi']}

    _param_validator = lambda p: (p['revival_time'] == 0 * u.ms and p['progenitor_mass'] == 30 * u.Msun
                                  and p['metallicity'] == 0.004) or \
                                 (p['revival_time'] != 0 * u.ms and p['eos'] == 'shen'
                                  and not (p['progenitor_mass'] == 30 * u.Msun and p['metallicity'] == 0.004))

    @_warn_deprecated_filename_argument
    def __new__(cls, filename=None, *, progenitor_mass=None, revival_time=None, metallicity=None, eos=None):
        """Model initialization.

        Parameters
        ----------
        filename : str
            Absolute or relative path to FITS file with model data. This argument is deprecated.

        Other Parameters
        ----------------
        progenitor_mass: astropy.units.Quantity
            Mass of model progenitor in units Msun. Valid values are {progenitor_mass}.
        revival_time: astropy.units.Quantity
            Time of shock revival in model in units ms. Valid values are {revival_time}.
            Selecting 0 ms will load a black hole formation model
        metallicity: float
            Progenitor metallicity. Valid values are {metallicity}.
        eos: str
            Equation of state. Valid values are {eos}.

        Raises
        ------
        ValueError
            If a combination of parameters is invalid when loading from parameters

        Examples
        --------
        >>> from snewpy.models.ccsn import Nakazato_2013; import astropy.units as u
        >>> Nakazato_2013(progenitor_mass=30*u.Msun, metallicity=0.004, revival_time=0*u.s, eos='togashi')
        Nakazato_2013 Model: nakazato-togashi-BH-z0.004-s30.0.fits
        Progenitor mass  : 30.0 solMass
        EOS              : togashi
        Metallicity      : 0.004
        Revival time     : 0.0 ms
        """
        # Attempt to load model from parameters
        if filename is not None:
            metadata = {'Progenitor mass': float(filename.split('-')[-1].strip('s%.fits')) * u.Msun}
            if 't_rev' in filename:
                metadata.update({
                    'EOS': filename.split('-')[-4].upper(),
                    'Metallicity': float(filename.split('-')[-3].strip('z%')),
                    'Revival time': float(filename.split('-')[-2].strip('t_rev%ms')) * u.ms
                })
            # No revival time because the explosion "failed" (BH formation).
            else:
                metadata.update({
                    'EOS': filename.split('-')[-4].upper(),
                    'Metallicity': float(filename.split('-')[-2].strip('z%')),
                    'Revival time': 0 * u.ms
                })
            return loaders.Nakazato_2013(os.path.abspath(filename), metadata)

        # Load from model parameters
        user_params = dict(zip(cls.param.keys(), (progenitor_mass, revival_time, metallicity, eos)))
        cls.check_valid_params(cls, **user_params)

        # Store model metadata.
        metadata = {
            'Progenitor mass': progenitor_mass,
            'EOS': eos,
            'Metallicity': metallicity,
            'Revival time': revival_time
        }

        # Strip units for filename construction
        progenitor_mass = progenitor_mass.to(u.Msun).value
        revival_time = revival_time.to(u.ms).value

        if revival_time != 0:
            filename = f"nakazato-{eos}-z{metallicity}-t_rev{int(revival_time)}ms-s{progenitor_mass:3.1f}.fits"
        else:
            filename = f"nakazato-{eos}-BH-z{metallicity}-s{progenitor_mass:3.1f}.fits"

        return loaders.Nakazato_2013(filename, metadata)

    # Populate Docstring with param values
    __new__.__doc__ = __new__.__doc__.format(**param)


class Sukhbold_2015(_RegistryModel):
    """Model based on simulations from Sukhbold et al., ApJ 821:38,2016. Models were shared privately by email.
    """
    param = {'progenitor_mass': [27., 9.6] * u.Msun,
             'eos': ['LS220', 'SFHo']}

    @_warn_deprecated_filename_argument
    def __new__(cls, filename=None, *, progenitor_mass=None, eos=None):
        """Model Initialization

        Parameters
        ----------
        filename : str
            Absolute or relative path to FITS file with model data. This argument is deprecated.

        Other Parameters
        ----------------
        progenitor_mass: astropy.units.Quantity
            Mass of model progenitor in units Msun. Valid values are {progenitor_mass}.
        eos: str
            Equation of state. Valid values are {eos}.

        Raises
        ------
        FileNotFoundError
            If a file for the chosen model parameters cannot be found
        ValueError
            If a combination of parameters is invalid when loading from parameters
        """
        if filename is not None:
            metadata = {
                'Progenitor mass': float(filename.split('-')[-1].strip('z%.fits')) * u.Msun,
                'EOS': filename.split('-')[-2]
            }
            return loaders.Sukhbold_2015(os.path.abspath(filename), metadata)

        user_params = dict(zip(cls.param.keys(), (progenitor_mass, eos)))
        cls.check_valid_params(cls, **user_params)

        if progenitor_mass.value == 9.6:
            filename = f'sukhbold-{eos}-z{progenitor_mass.value:3.1f}.fits'
        else:
            filename = f'sukhbold-{eos}-s{progenitor_mass.value:3.1f}.fits'

        metadata = {
            'Progenitor mass': progenitor_mass,
            'EOS': eos
        }
        return loaders.Sukhbold_2015(filename, metadata)

    # Populate Docstring with param values
    __new__.__doc__ = __new__.__doc__.format(**param)


class Tamborra_2014(_RegistryModel):
    """Model based on 3D simulations from `Tamborra et al., PRD 90:045032, 2014 <https://arxiv.org/abs/1406.0006>`_.
    Data files are from the `Garching Supernova Archive`_.
    """

    param = {'progenitor_mass': [11.2, 20., 27.] * u.Msun,
             'direction': [1,2,3]}
    _param_validator = lambda p: (p['progenitor_mass'] == 11.2 * u.Msun and p['direction'] in (1,)) or \
        (p['progenitor_mass'] == 20. * u.Msun and p['direction'] in (1,)) or \
        (p['progenitor_mass'] == 27. * u.Msun and p['direction'] in (1,2,3))
    
    @_warn_deprecated_filename_argument
    def __new__(cls, filename=None, eos='LS220', *, progenitor_mass=None, direction=None):
        if filename is not None:
            # Metadata creation is implemented in snewpy.models.base._GarchingArchiveModel
            return loaders.Tamborra_2014(os.path.abspath(filename))

        cls.check_valid_params(cls, progenitor_mass=progenitor_mass, direction=direction)
        filename = f's{progenitor_mass.value:3.1f}c_3D_dir{direction}'

        metadata = {
            'Progenitor mass': progenitor_mass,
            'EOS': 'LS220'
        }

        # Metadata is handled by __init__ in _GarchingArchiveModel
        return loaders.Tamborra_2014(filename=filename, metadata=metadata)

    # Populate Docstring with param values
    __new__.__doc__ = loaders.Tamborra_2014.__init__.__doc__.format(**param)


class Bollig_2016(_RegistryModel):
    """Model based on simulations from `Bollig et al. (2016) <https://arxiv.org/abs/1508.00785>`_.
    Models were taken, with permission, from the `Garching Supernova Archive`_.
    """

    param = {'progenitor_mass': [11.2, 27.] * u.Msun}

    def __new__(cls, filename=None,  eos='LS220', *, progenitor_mass=None):
        if filename is not None:
            # Metadata creation is implemented in snewpy.models.base._GarchingArchiveModel
            return loaders.Bollig_2016(os.path.abspath(filename))

        cls.check_valid_params(cls, progenitor_mass=progenitor_mass)
        filename = f's{progenitor_mass.value:3.1f}c'

        metadata = {
            'Progenitor mass': progenitor_mass,
            'EOS': 'LS220'
        }

        return loaders.Bollig_2016(filename=filename, metadata=metadata)

    # Populate Docstring with param values (Docstring is inherited from base._GarchingArchiveModel.__init__)
    __new__.__doc__ = loaders.Bollig_2016.__init__.__doc__.format(**param)


class Walk_2018(_RegistryModel):
    """Model based on SASI-dominated simulations from `Walk et al.,
    PRD 98:123001, 2018 <https://arxiv.org/abs/1807.02366>`_. Data files are from
    the `Garching Supernova Archive`_.
    """

    param = {'progenitor_mass': 15. * u.Msun, 'rotation': ['fast','slow','non'], 'direction': [1,2,3]}

    @_warn_deprecated_filename_argument
    def __new__(cls, filename=None, eos='LS220', *, progenitor_mass=None, rotation=None, direction=None):
        if filename is not None:
            # Metadata creation is implemented in snewpy.models.base._GarchingArchiveModel
            return loaders.Walk_2018(os.path.abspath(filename))

        cls.check_valid_params(cls, progenitor_mass=progenitor_mass, rotation=rotation, direction=direction)
        filename = f's{progenitor_mass.value:3.1f}c_3D_{rotation}rot_dir{direction}'

        metadata = {
            'Progenitor mass': progenitor_mass,
            'EOS': 'LS220'
        }

        return loaders.Walk_2018(filename=filename, metadata=metadata)

    # Populate Docstring with param values (Docstring is inherited from base._GarchingArchiveModel.__init__)
    __new__.__doc__ = loaders.Walk_2018.__init__.__doc__.format(**param)


class Walk_2019(_RegistryModel):
    """Model based on SASI-dominated simulations from `Walk et al.,
    PRD 101:123013, 2019 <https://arxiv.org/abs/1910.12971>`_. Data files are
    from the `Garching Supernova Archive`_.
    """

    param = {'progenitor_mass': [40., 75.] * u.Msun,
             'direction': [1,2,3]}
    _param_validator = lambda p: (p['progenitor_mass'] == 75. * u.Msun and p['direction'] in (1,2)) or \
        (p['progenitor_mass'] == 40. * u.Msun and p['direction'] in (1,2,3))
    
    @_warn_deprecated_filename_argument
    def __new__(cls, filename=None, eos='LS220', *, progenitor_mass=None, direction=None):
        if filename is not None:
            # Metadata creation is implemented in snewpy.models.base._GarchingArchiveModel
            return loaders.Walk_2019(os.path.abspath(filename))

        cls.check_valid_params(cls, progenitor_mass=progenitor_mass, direction=direction)
        filename = f's{progenitor_mass.value:3.1f}c_3DBH_dir{direction}'

        metadata = {
            'Progenitor mass': progenitor_mass,
            'EOS': 'LS220'
        }

        return loaders.Walk_2019(filename=filename, metadata=metadata)

    # Populate Docstring with param values (Docstring is inherited from base._GarchingArchiveModel.__init__)
    __new__.__doc__ = loaders.Walk_2019.__init__.__doc__.format(**param)


class OConnor_2013(_RegistryModel):
    """Model based on the black hole formation simulation in `O'Connor & Ott (2013) <https://arxiv.org/abs/1207.1100>`_.
    """

    param = {'progenitor_mass': (list(range(12, 34)) +
                                 list(range(35, 61, 5)) +
                                 [70, 80, 100, 120]) * u.Msun,
             'eos': ['HShen', 'LS220']}

    _param_abbrv = {'progenitor_mass': '[12..33, 35..5..60, 70, 80, 100, 120] solMass',
                    'eos': ['HShen', 'LS220']}

    @_warn_deprecated_filename_argument
    def __new__(cls, base=None, mass=None, eos='LS220', *, progenitor_mass=None):
        """Model Initialization.

        Parameters
        ----------
        base : str
            Absolute or relative path folder with model data. This argument is deprecated.
        mass: int
            Mass of model progenitor in units Msun. This argument is deprecated.
        eos: str
            Equation of state. Valid values are {eos}.

        Other Parameters
        ----------------
        progenitor_mass: astropy.units.Quantity
            Mass of model progenitor in units Msun. Valid values are {progenitor_mass}.

        Raises
        ------
        FileNotFoundError
            If a file for the chosen model parameters cannot be found
        ValueError
            If a combination of parameters is invalid when loading from parameters

        """
        # TODO: (For v2.0) Change `base` to filename, move compressed model files to OConnor_2013 model folder
        if mass is not None:
            warn(f'Argument `mass` of type int is deprecated. To initialize this model, use keyword arguments '
                 f'{list(cls.param.keys())}. See {cls.__name__}.param, {cls.__name__}.get_param_combinations() for more info',
                 category=DeprecationWarning, stacklevel=2)
        else:
            mass = 15  # Default Value, this is handled this way for backwards compatibility -- TODO (For V2.0) Remove

        if base is not None:
            # If base is provided, do not attempt to load from param.
            if mass * u.Msun not in cls.param['progenitor_mass']:
                raise ValueError(f'Invalid value for argument `progenitor mass` or `mass`, see {cls.__name__}.param'
                                 f' for allowed values')
            metadata = {'Progenitor mass': progenitor_mass if progenitor_mass is not None else mass * u.Msun,
                        'EOS': eos}

            filename = os.path.join(base, f"{eos}_timeseries.tar.gz")
            return loaders.OConnor_2013(os.path.abspath(filename), metadata)

        # Load from Parameters
        cls.check_valid_params(cls, progenitor_mass=progenitor_mass, eos=eos)
        filename = f'{eos}_timeseries.tar.gz'

        metadata = {
            'Progenitor mass': progenitor_mass,
            'EOS': eos,
        }

        return loaders.OConnor_2013(filename=filename, metadata=metadata)

    # Populate Docstring with param values
    __new__.__doc__ = __new__.__doc__.format(**_param_abbrv)


class OConnor_2015(_RegistryModel):
    """Model based on the black hole formation simulation in `O'Connor (2015) <https://arxiv.org/abs/1411.7058>`_.
    """

    param = {'progenitor_mass': 40 * u.Msun}

    @_warn_deprecated_filename_argument
    def __new__(cls, filename=None, eos='LS220', *, progenitor_mass=None):
        """Model Initialization.

        Parameters
        ----------
        filename : str
            Absolute or relative path to tar.gz file with model data. This argument is deprecated.
        eos: str
            Equation of state. Valid value is 'LS220'. This argument is deprecated.

        Other Parameters
        ----------------
        progenitor_mass: astropy.units.Quantity
            Mass of model progenitor in units Msun. Valid values are {progenitor_mass}.

        Raises
        ------
        FileNotFoundError
            If a file for the chosen model parameters cannot be found
        ValueError
            If a combination of parameters is invalid when loading from parameters
        """
        metadata = {
            'Progenitor mass': 40*u.Msun,
            'EOS': 'LS220',
        }

        if filename is not None:
            return loaders.OConnor_2015(os.path.abspath(filename), metadata)

        # Load from Parameters
        cls.check_valid_params(cls, progenitor_mass=progenitor_mass)
        # Filename is currently the same regardless of parameters
        filename = 'M1_neutrinos.dat'

        return loaders.OConnor_2015(filename, metadata)

    # Populate Docstring with param values
    __new__.__doc__ = __new__.__doc__.format(**param)


class Zha_2021(_RegistryModel):
    """Model based on the hadron-quark phse transition models from `Zha et al. 2021 <https://arxiv.org/abs/2103.02268>`_.
    """

    param = {'progenitor_mass': (list(range(16, 27)) + [19.89, 22.39, 30, 33]) * u.Msun}

    _param_abbrv = {'progenitor_mass': '[16..26, 19.89, 22.39, 30, 33] solMass'}

    @_warn_deprecated_filename_argument
    def __new__(cls, filename=None, eos='ST0S_B145', *, progenitor_mass=None):
        """Model Initialization.

        Parameters
        ----------
        filename : str
            Absolute or relative path to file with model data. This argument is deprecated.
        eos : str
            Equation of state. Valid value is 'ST0S_B145'. This argument is deprecated.

        Other Parameters
        ----------------
        progenitor_mass: astropy.units.Quantity
            Mass of model progenitor in units Msun. Valid values are {progenitor_mass}.

        Raises
        ------
        FileNotFoundError
            If a file for the chosen model parameters cannot be found
        ValueError
            If a combination of parameters is invalid when loading from parameters
        """
        if filename is not None:
            metadata = {'Progenitor mass': float(os.path.splitext(os.path.basename(filename))[0][1:]) * u.Msun,
                        'EOS': 'STOS_B145'}
            return loaders.Zha_2021(os.path.abspath(filename), metadata)

        # Load from Parameters
        cls.check_valid_params(cls, progenitor_mass=progenitor_mass)

        metadata = {
            'Progenitor mass': progenitor_mass,
            'EOS': 'STOS_B145',
        }

        filename = f's{progenitor_mass.value:g}.dat'

        return loaders.Zha_2021(filename, metadata)

    # Populate Docstring with abbreviated param values
    __new__.__doc__ = __new__.__doc__.format(**_param_abbrv)


class Warren_2020(_RegistryModel):
    """Model based on simulations from Warren et al., ApJ 898:139, 2020.
    Neutrino fluxes available at https://doi.org/10.5281/zenodo.3667908."""
    # TODO: (For v2.0) Resolve Zenodo issues (Missing files)
    # np.arange with decimal increments can produce floating point errors
    # Though it may be more intuitive to use np.arange, these fp-errors quickly become troublesome
    param = {'progenitor_mass': np.concatenate((np.linspace(9.0, 12.75, 16),
                                                np.linspace(13, 30., 171),
                                                np.linspace(31., 33., 3),
                                                np.linspace(35, 55, 5),
                                                np.linspace(60, 80, 3),
                                                np.linspace(100, 120, 2))) * u.Msun,
             'turbmixing_param': [1.23, 1.25, 1.27]}

    _param_abbrv = {'progenitor_mass': '[9..0.25..13, 13..0.1..30, 31..35, 35..5..60, 70..10..90, 100, 120] solMass',
                    'turbmixing_param': [1.23, 1.25, 1.27]}


    @_warn_deprecated_filename_argument
    def __new__(cls, filename=None, eos='SFHo', *, progenitor_mass=None, turbmixing_param=None):
        """
        Parameters
        ----------
        filename : str
            Absolute or relative path to file with model data. This argument is deprecated.
        eos : str
            Equation of state. Valid value is 'SFHo'. This argument is deprecated.

        Other Parameters
        ----------------
        progenitor_mass: astropy.units.Quantity
            Mass of model progenitor in units Msun. Valid values are {progenitor_mass}.
        turbmixing_param: float
            Turbulent mixing parameter alpha_lambda. Valid Values are {turbmixing_param}

        Raises
        ------
        FileNotFoundError
            If a file for the chosen model parameters cannot be found
        ValueError
            If a combination of parameters is invalid when loading from parameters
        """
        if filename is not None:
            _, _, turbmixing_param, progenitor_mass = os.path.splitext(os.path.basename(filename))[0].split('_')
            metadata = {'Progenitor mass': float(progenitor_mass[1:]) * u.Msun,
                        'Turb. mixing param.': float(turbmixing_param[1:]),
                        'EOS': 'SFHo'}

            return loaders.Warren_2020(os.path.abspath(filename), metadata)

        # Load from Parameters
        user_params = dict(zip(cls.param.keys(), (progenitor_mass, turbmixing_param)))
        cls.check_valid_params(cls, **user_params)

        if progenitor_mass.value.is_integer() and progenitor_mass.value <= 30.:
            fname = os.path.join(f'stir_a{turbmixing_param:3.2f}',
                                 f'stir_multimessenger_a{turbmixing_param:3.2f}_m{progenitor_mass.value:.1f}.h5')
        else:
            fname = os.path.join(f'stir_a{turbmixing_param:3.2f}',
                                 f'stir_multimessenger_a{turbmixing_param:3.2f}_m{progenitor_mass.value:g}.h5')

        # Set model metadata.
        metadata = {
            'Progenitor mass': progenitor_mass,
            'Turb. mixing param.': turbmixing_param,
            'EOS': 'SFHo',
        }

        return loaders.Warren_2020(fname, metadata)

    # Populate Docstring with abbreviated param values
    __new__.__doc__ = __new__.__doc__.format(**_param_abbrv)


class Kuroda_2020(_RegistryModel):
    """Model based on simulations from `Kuroda et al. (2020) <https://arxiv.org/abs/2009.07733>`_."""

    param = {'rotational_velocity': [0, 1] * u.rad / u.s,
             'magnetic_field_exponent': [0, 12, 13]}
    _param_validator = lambda p: (p['rotational_velocity'].value == 1 and p['magnetic_field_exponent'] in (12, 13)) or \
                               (p['rotational_velocity'].value == 0 and p['magnetic_field_exponent'] == 0)

    @_warn_deprecated_filename_argument
    def __new__(cls, filename=None, eos='LS220', mass=20*u.Msun, *, rotational_velocity=None,
                magnetic_field_exponent=None):
        """
        Parameters
        ----------
        filename : str
            Absolute or relative path to file with model data. This argument is deprecated.
        eos : str
            Equation of state. Valid value is 'LS220'. This argument is deprecated.
        mass: astropy.units.Quantity
            Mass of model progenitor in units Msun. Valid value is 20 * u.Msun. This argument is deprecated.

        Other Parameters
        ----------------
        rotational_velocity: astropy.units.Quantity
            Rotational velocity of progenitor. Valid values are {rotational_velocity}
        magnetic_field_exponent: int
            Exponent of magnetic field (See Eq. 46). Valid Values are {magnetic_field_exponent}

        Raises
        ------
        FileNotFoundError
            If a file for the chosen model parameters cannot be found
        ValueError
            If a combination of parameters is invalid when loading from parameters
        """
        if filename is not None:
            _, rotational_velocity, magnetic_field_exponent = re.split('R|B',
                                                                       os.path.splitext(os.path.basename(filename))[0])
            metadata = {
                'Progenitor mass': 20 * u.Msun,
                'Rotational Velocity': int(rotational_velocity[0]),
                'B_0 Exponent': int(magnetic_field_exponent),
                'EOS': 'LS220'
            }
            return loaders.Kuroda_2020(os.path.abspath(filename), metadata)

        # Load from Parameters
        cls.check_valid_params(cls, rotational_velocity=rotational_velocity,
                               magnetic_field_exponent=magnetic_field_exponent)
        filename = f'LnuR{int(rotational_velocity.value):1d}0B{int(magnetic_field_exponent):02d}.dat'

        metadata = {
            'Progenitor mass': 20 * u.Msun,
            'Rotational Velocity': rotational_velocity,
            'B_0 Exponent': magnetic_field_exponent,
            'EOS': 'LS220',
            }

        return loaders.Kuroda_2020(filename, metadata)

    __new__.__doc__ = __new__.__doc__.format(**param)


class Fornax_2019(_RegistryModel):
    """Model based on 3D simulations from D. Vartanyan, A. Burrows, D. Radice, M.  A. Skinner and J. Dolence, MNRAS 482(1):351, 2019. 
       Data available at https://www.astro.princeton.edu/~burrows/nu-emissions.3d/
    """
    param = {'progenitor_mass': [9, 10, 12, 13, 14, 15, 16, 19, 25, 60] * u.Msun}

    @_warn_deprecated_filename_argument
    def __new__(cls, filename=None, cache_flux=False, *, progenitor_mass=None, ):
        """Model Initialization.

        Parameters
        ----------
        filename : str
            Absolute or relative path to file with model data. This argument is deprecated.
        cache_flux : bool
            If true, pre-compute the flux on a fixed angular grid and store the values in a FITS file.

        Other Parameters
        ----------------
        progenitor_mass: astropy.units.Quantity
            Mass of model progenitor in units Msun. Valid values are {progenitor_mass}.
        """
        if filename is not None:
            progenitor_mass = os.path.splitext(os.path.basename(filename))[0].split('_')[2]
            metadata = {'Progenitor mass': int(progenitor_mass[:-1]) * u.Msun}
            return loaders.Fornax_2019(os.path.abspath(filename), metadata, cache_flux=cache_flux)

        # Load from Parameters
        metadata = {'Progenitor mass': progenitor_mass}

        cls.check_valid_params(cls, progenitor_mass=progenitor_mass)
        if progenitor_mass.value == 16:
            filename = f'lum_spec_{int(progenitor_mass.value):d}M_r250.h5'
        else:
            filename = f'lum_spec_{int(progenitor_mass.value):d}M.h5'

        return loaders.Fornax_2019(filename, metadata, cache_flux=cache_flux)

    # Populate Docstring with abbreviated param values
    __new__.__doc__ = __new__.__doc__.format(**param)


class Fornax_2021(_RegistryModel):
    """Model based on 3D simulations from D. Vartanyan, A. Burrows, D. Radice, M.  A. Skinner and J. Dolence, MNRAS 482(1):351, 2019. 
       Data available at https://www.astro.princeton.edu/~burrows/nu-emissions.3d/
        """
    param = {'progenitor_mass': (list(range(12, 24)) + [25, 26, 26.99]) * u.Msun}

    _param_abbrv = {'progenitor_mass': '[12..26, 26.99] solMass'}

    @_warn_deprecated_filename_argument
    def __new__(cls, filename=None, *, progenitor_mass=None):
        """Model Initialization.

        Parameters
        ----------
        filename : str
            Absolute or relative path to file with model data. This argument is deprecated.

        Other Parameters
        ----------------
        progenitor_mass: astropy.units.Quantity
            Mass of model progenitor in units Msun. Valid values are {progenitor_mass}.
        """
        if filename is not None:
            progenitor_mass = os.path.splitext(os.path.basename(filename))[0].split('_')[2]
            metadata = {'Progenitor mass': float(progenitor_mass[:-1]) * u.Msun}
            return loaders.Fornax_2021(os.path.abspath(filename), metadata)

        # Load from Parameters
        cls.check_valid_params(cls, progenitor_mass=progenitor_mass)
        if progenitor_mass.value.is_integer():
            filename = f'lum_spec_{int(progenitor_mass.value):2d}M_r10000_dat.h5'
        else:
            filename = f'lum_spec_{progenitor_mass.value:.2f}M_r10000_dat.h5'

        metadata = {'Progenitor mass': progenitor_mass}

        return loaders.Fornax_2021(filename, metadata)

    # Populate Docstring with abbreviated param values
    __new__.__doc__ = __new__.__doc__.format(**_param_abbrv)

class SNOwGLoBES:
    """A model that does not inherit from SupernovaModel (yet) and imports a group of SNOwGLoBES files."""

    def __init__(self, tarfilename):
        """
        Parameters
        ----------
        tarfilename: str
            Absolute or relative path to tar archive with SNOwGLoBES files.
        """
        self.tfname = tarfilename
        tf = tarfile.open(self.tfname)

        # For now just pull out the "NoOsc" files.
        datafiles = sorted([f.name for f in tf if '.dat' in f.name])
        noosc = [df for df in datafiles if 'NoOsc' in df]
        noosc.sort(key=len)

        # Loop through the noosc files and pull out the number fluxes.
        self.time = []
        self.energy = None
        self.flux = {}
        self.fmin = 1e99
        self.fmax = -1e99

        for nooscfile in noosc:
            with tf.extractfile(nooscfile) as f:
                logging.debug('Reading {}'.format(nooscfile))
                meta = f.readline()
                metatext = meta.decode('utf-8')
                t = float(metatext.split('TBinMid=')[-1].split('sec')[0])
                dt = float(metatext.split('tBinWidth=')[-1].split('s')[0])
                dE = float(metatext.split('eBinWidth=')[-1].split('MeV')[0])

                data = Table.read(f, format='ascii.commented_header', header_start=-1)
                data.meta['t'] = t
                data.meta['dt'] = dt
                data.meta['dE'] = dE

                self.time.append(t)
                if self.energy is None:
                    self.energy = (data['E(GeV)'].data*1000).tolist()

            for flavor in ['NuE', 'NuMu', 'NuTau', 'aNuE', 'aNuMu', 'aNuTau']:
                if flavor in self.flux:
                    self.flux[flavor].append(data[flavor].data.tolist())
                else:
                    self.flux[flavor] = [data[flavor].data.tolist()]

        # We now have a table with rows=times and columns=energies. Transpose
        # so that rows=energy and cols=time.
        for k, v in self.flux.items():
            self.flux[k] = np.transpose(self.flux[k])
            self.fmin = np.minimum(self.fmin, np.min(self.flux[k]))
            self.fmax = np.maximum(self.fmax, np.max(self.flux[k]))

    def get_fluence(self, t):
        """Return the fluence at a given time t.

        Parameters
        ----------
        t : float
            Time in seconds.

        Returns
        -------
        fluence : dict
            A dictionary giving fluence at time t, keyed by flavor.
        """
        # Get index of closest element in the array
        idx = np.abs(np.asarray(self.time) - t).argmin()

        fluence = {}
        for k, fl in self.flux.items():
            fluence[k] = fl[:,idx]

        return fluence



import pytest
from snewpy.snowglobes import generate_fluence, SimpleRate
from pathlib import Path
import tarfile

pytestmark=pytest.mark.snowglobes

models = ['Bollig_2016/s11.2c', 'Bollig_2016/s27.0c']
transformations = ['AdiabaticMSW_NMO','AdiabaticMSW_IMO']

@pytest.mark.parametrize('model',models)
@pytest.mark.parametrize('transformation',transformations)
def test_generate_fluence(splr, model, transformation):
    """make some input fluxes"""
    model_path = Path('./models')/model
    model_type = model_path.parent.stem
    output_name = f'fluence_{model.replace("/","_")}_{transformation}'
    fname = generate_fluence(model_path,model_type, transformation, d=10, output_filename=output_name)
    with tarfile.open(fname) as tar:
        tar.extractall(model_path.parent)

#numbers from SNEWSv2 paper - for approximate checks 
crosscheck_table=[  ('icecube',          320e3),
                    ('wc100kt30prct',4044.8/0.32),
                    ('ar40kt',              2700),
                    ('novaND',                40),
                    ('novaFD',              1900),
                    ('halo1',                  4),
                    ('halo2',                 53)]

#numbers from the SNEWPY SNOwGLoBES interface - for approximate checks 
#of simple rate computation (no smearing)
crosscheck_table_unsmeared=[  ('icecube',          7.2352e6),
                    ('wc100kt30prct',4486.9/0.32),
                    ('ar40kt',              2858.7),
                    ('novaND',                41.7231),
                    ('novaFD',              1947.0768),
                    ('halo1',                  11.5709),
                    ('halo2',                 146.4671)]


@pytest.fixture(autouse=True)
def splr():
    return SimpleRate()

@pytest.mark.parametrize('detector, expected_total_unsmeared',crosscheck_table_unsmeared)
def test_simplerate_crosscheck(splr, detector, expected_total_unsmeared):
    flux = './models/Bollig_2016/fluence_Bollig_2016_s11.2c_AdiabaticMSW_NMO.dat'
    data = splr.run(flux,detector)
    total = data[0].weighted.unsmeared.sum().sum()
    assert total == pytest.approx(expected_total_unsmeared, 0.01)

@pytest.mark.parametrize('detector, expected_total',crosscheck_table)
def test_simplerate_smear_crosscheck(splr, detector, expected_total):
    flux = './models/Bollig_2016/fluence_Bollig_2016_s11.2c_AdiabaticMSW_NMO.dat'
    data = splr.run(flux,detector)
    total = data[0].weighted.smeared.sum().sum()
    assert total == pytest.approx(expected_total, 0.1)



import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
from astropy import units as u
from .base import SupernovaModel
from snewpy.neutrino import Flavor
from pathlib import Path

def _interp_T(t0, v0, dt=1e-3, dv=1e-10, axis=0):
    "dilog interpolation"
    lt0 = np.log(t0 + dt)
    lv0 = np.log(v0 + dv)
    lv1 = interp1d(lt0, lv0, axis=axis, fill_value=0, copy=False, bounds_error=False)
    return lambda t1: np.exp(lv1(np.log(t1 + dt)))


def _interp_E(e0, v0, axis=1):
    "linear interpolation"
    return interp1d(e0, v0, axis=axis, fill_value=0, copy=False, bounds_error=False)


def _interp_TE(times, energies, array, ax_t=1, ax_e=2):
    def _f(t, E):
        a_t = _interp_T(times, array, axis=ax_t)(t)
        a_te = _interp_E(energies, a_t, axis=ax_e)(E)
        return a_te

    return _f

class Odrzywolek_2010(SupernovaModel):
    """Set up a presupernova model, based on 
    [A. Odrzywolek and A. Heger, Acta Phys. Polon. B 41 (2010) 1611.]
    """

    def __init__(self, fname):
        df = pd.read_csv(
            fname,
            delim_whitespace=True,
            skiprows=1,
            usecols=[1,6,7,8],
            names=["time","a","alpha","b"],
            index_col="time",
        )
        # interpolated in time
        self.df_t = _interp_T(df.index, df)
        self.factor = {}
        for f in Flavor:
            if f.is_electron:
                self.factor[f] = 1.0
            else:
                # nuX/nuE ratio from Odrzywolek paper: (arXiv:astro-ph/0311012)
                self.factor[f] = 0.36
        self.filename = Path(fname).stem
        self.progenitor_mass = float(self.filename.split('_')[0][1:]) * u.Msun
        time = -df.index.to_numpy() << u.s
        metadata = {'Progenitor Mass': self.progenitor_mass}
        super().__init__(time, metadata)

    def get_initial_spectra(self, t, E, flavors=Flavor):
        # negative t for time before SN
        t = -t.to_value("s")
        E = E.to_value("MeV")
        df = self.df_t(t)
        a, alpha, b = df.T
        Enu = np.expand_dims(E, 1)
        a = np.expand_dims(a, 0)
        alpha = np.expand_dims(alpha, 0)
        b = np.expand_dims(b, 0)
        fluence = a * Enu ** alpha * np.exp(-b * Enu) / (u.MeV * u.s)
        result = {f: fluence.T * self.factor[f] for f in flavors}
        return result


class Patton_2017(SupernovaModel):
    """Set up a presupernova model based on 
    [Kelly M. Patton et al 2017 ApJ 851 6, 
     https://doi.org/10.5281/zenodo.2598709]
    """
    def __init__(self, fname):
        df = pd.read_csv(
            fname,
            comment="#",
            delim_whitespace=True,
            names=["time","Enu",Flavor.NU_E,Flavor.NU_E_BAR,Flavor.NU_X,Flavor.NU_X_BAR],
            usecols=range(6),
        )

        df = df.set_index(["time", "Enu"])
        times = df.index.levels[0].to_numpy()
        energies = df.index.levels[1].to_numpy()
        df = df.unstack("Enu")
        # make a 3d array
        self.array = np.stack([df[f] for f in Flavor], axis=0)
        self.interpolated = _interp_TE(
            times, energies, self.array, ax_t=1, ax_e=2
        )
        self.filename = Path(fname).stem
        self.progenitor_mass = float(self.filename.split('_')[-1].rstrip('SolarMass')) * u.Msun
        metadata = {'Progenitor Mass': self.progenitor_mass}
        super().__init__(times << u.hour, metadata)

    def get_initial_spectra(self, t, E, flavors=Flavor):
        t = np.array(-t.to_value("hour"), ndmin=1)
        E = np.array(E.to_value("MeV"), ndmin=1)
        flux = self.interpolated(t, E) / (u.MeV * u.s)
        return {f: flux[f] for f in flavors}

class Kato_2017(SupernovaModel):
    """Set up a presupernova model based on 
    [Chinami Kato et al 2017 ApJ 848 48]
    """
    def __init__(self, path):
        fluxes = {}
        #reading the time steps values:
        times, step = np.loadtxt(f"{path}/total_nue/lightcurve_nue_all.dat", usecols=[0, 3]).T

        file_base = {Flavor.NU_E: 'total_nue/spe_all',
                 Flavor.NU_E_BAR:'total_nueb/spe_all',
                 Flavor.NU_X:    'total_nux/spe_sum_mu_nu',
                 Flavor.NU_X_BAR:'total_nux/spe_sum_mu'
                 }
        for flv,file_base in file_base.items():
            d2NdEdT = []
            for s in step:
                energies, dNdE = np.loadtxt(f"{path}/{file_base}{s:05.0f}.dat").T
                d2NdEdT += [dNdE]
            fluxes[flv] = np.stack(d2NdEdT)
        self.array = np.stack([fluxes[f] for f in Flavor], axis=0)
        self.interpolated = _interp_TE(
            times, energies, self.array, ax_t=1, ax_e=2
        )
        self.filename = Path(path).stem
        self.progenitor_mass = float(self.filename.strip('m')) * u.Msun
        metadata = {'Progenitor Mass': self.progenitor_mass}
        super().__init__(times << u.s, metadata)

    def get_initial_spectra(self, t, E, flavors=Flavor):
        t = np.array(-t.to_value("s"), ndmin=1)
        E = np.array(E.to_value("MeV"), ndmin=1)
        flux = self.interpolated(t, E) / (u.MeV * u.s)
        return {f: flux[f] for f in flavors}




# Git interaction code adapted from https://github.com/desihub/desiutil.
# The desiutil project is distributed under a 3-clause BSD style license:
#
# Copyright (c) 2014-2017, DESI Collaboration <desi-data@desi.lbl.gov>
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of the DESI Collaboration nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
"""Some code for interacting with git.
"""

import re
from os.path import abspath, isdir, isfile, join
from setuptools import Command
from distutils.log import INFO


def get_version():
    """Get the value of ``__version__`` without having to import the module.

    Returns
    -------
    :class:`str`
        The value of ``__version__``.
    """
    ver = 'unknown'
    try:
        version_dir = find_version_directory()
    except IOError:
        return ver
    version_file = join(version_dir, '_version.py')
    if not isfile(version_file):
        update_version()
    with open(version_file, "r") as f:
        for line in f.readlines():
            mo = re.match("__version__ = '(.*)'", line)
            if mo:
                ver = mo.group(1)
    return ver


def get_git_version(git='git'):
    """Use ``git describe`` to generate a version string.

    Parameters
    ----------
    git : :class:`str`, optional
        Path to the git executable, if not in :envvar:`PATH`.

    Returns
    -------
    :class:`str`
        A :pep:`386`-compatible version string.

    Notes
    -----
    The version string should be compatible with :pep:`386` and
    :pep:`440`.
    """
    from subprocess import Popen, PIPE
    myversion = '0.0.1.dev0'
    try:
        p = Popen([git, "describe", "--tags", "--dirty", "--always"],
                  universal_newlines=True, stdout=PIPE, stderr=PIPE)
    except OSError:
        return myversion
    out, err = p.communicate()
    if p.returncode != 0:
        return myversion
    ver = out.rstrip().split('-')[0]+'.dev'
    try:
        p = Popen([git, "rev-list", "--count", "HEAD"],
                  universal_newlines=True, stdout=PIPE, stderr=PIPE)
    except OSError:
        return myversion
    out, err = p.communicate()
    if p.returncode != 0:
        return myversion
    ver += out.rstrip()
    return ver


def update_version(tag=None):
    """Update the _version.py file.

    Parameters
    ----------
    tag : :class:`str`, optional
        Set the version to this string, unconditionally.

    Raises
    ------
    IOError
        If the repository type could not be determined.
    """
    version_dir = find_version_directory()
    if tag is not None:
        ver = tag
    else:
        if isdir(".git"):
            ver = get_git_version()
        else:
            raise IOError("Repository type is not git.")
    version_file = join(version_dir, '_version.py')
    with open(version_file, "w") as f:
        f.write("__version__ = '{}'\n".format(ver))
    return


def find_version_directory():
    """Return the name of a directory containing version information.

    Looks for files in the following places:

    * python/snewpy/_version.py
    * snewpy/_version.py

    Returns
    -------
    :class:`str`
        Name of a directory that can or does contain version information.

    Raises
    ------
    IOError
        If no valid directory can be found.
    """
    packagename='snewpy'
    setup_dir = abspath('.')
    if isdir(join(setup_dir, 'python', packagename)):
        version_dir = join(setup_dir, 'python', packagename)
    elif isdir(join(setup_dir, packagename)):
        version_dir = join(setup_dir, packagename)
    else:
        raise IOError("Could not find a directory containing version information!")
    return version_dir


class SetVersion(Command):
    """Allow users to easily update the package version with
    ``python setup.py version``.
    """
    description = "update _version.py from git repo"
    user_options = [('tag=', 't',
                     'Set the version to a name in preparation for tagging.'),
                    ]
    boolean_options = []

    def initialize_options(self):
        self.tag = None

    def finalize_options(self):
        pass

    def run(self):
        update_version(tag=self.tag)
        ver = get_version()
        self.announce("Version is now {}.".format(ver), level=INFO)



# -*- coding: utf-8 -*-
"""Integration test based on SNEWS2.0_rate_table_singleexample.py
"""
import unittest
from snewpy import snowglobes


class TestSimpleRate(unittest.TestCase):

    def test_simplerate(self):
        """Integration test based on SNEWS2.0_rate_table_singleexample.py
        """
        # Hardcoded paths on GitHub Action runner machines
        SNOwGLoBES_path = None
        SNEWPY_model_dir = "models/"

        distance = 10  # Supernova distance in kpc
        detector = "wc100kt30prct" #SNOwGLoBES detector for water Cerenkov
        modeltype = 'Bollig_2016' # Model type from snewpy.models
        model = 's11.2c' # Name of model
        transformation = 'AdiabaticMSW_NMO' # Desired flavor transformation

        # Construct file system path of model file and name of output file
        model_path = SNEWPY_model_dir + "/" + modeltype + "/" + model
        outfile = modeltype + "_" + model + "_" + transformation

        # Now, do the main work:
        print("Generating fluence files ...")
        tarredfile = snowglobes.generate_fluence(model_path, modeltype, transformation, distance, outfile)

        print("Simulating detector effects with SNOwGLoBES ...")
        snowglobes.simulate(SNOwGLoBES_path, tarredfile, detector_input=detector, detector_effects=True)

        print("Collating results ...")
        tables = snowglobes.collate(SNOwGLoBES_path, tarredfile, skip_plots=True, smearing=True)

        # Use results to print the number of events in different interaction channels
        key = f"Collated_{outfile}_{detector}_events_unsmeared_weighted.dat"
        total_events = 0
        for i, channel in enumerate(tables[key]['header'].split()):
            if i == 0:
                continue
            n_events = sum(tables[key]['data'][i])
            total_events += n_events
            print(f"{channel:10}: {n_events:.3f} events")

        #Super-K has 32kT inner volume
        print("Total events in Super-K-like detector:" , 0.32*total_events)

        # Use results to print the number of events in different interaction channels
        # with efficiency and smearing
        key = f"Collated_{outfile}_{detector}_events_smeared_weighted.dat"
        total_events_smeared = 0
        for i, channel in enumerate(tables[key]['header'].split()):
            if i == 0:
                continue
            n_events = sum(tables[key]['data'][i])
            total_events_smeared += n_events
            print(f"{channel:10}: {n_events:.3f} events (smeared)")

        #Super-K has 32kT inner volume
        print("Total events in Super-K-like detector (with smearing):" , 0.32*total_events_smeared)

        # We do not use the SNOwGLoBES scaling factors but use other constants so we do not
        # expect the results to agree to 7 digits. Here sub-permille agreement is good enough.
        sk_expected = 4486.929197175579
        sk_expected_smeared = 4044.841743901513
        sk_computed = 0.32 * total_events
        sk_computed_smeared = 0.32 * total_events_smeared
        discrepancy = abs(sk_computed - sk_expected)/sk_expected
        discrepancy_smeared = abs(sk_computed_smeared - sk_expected_smeared)/sk_expected_smeared

        assert discrepancy < 0.001, f"Number of unsmeared events computed for SK is {sk_computed}, should be {sk_expected}"
        assert discrepancy_smeared < 0.001, f"Number of smeared events computed for SK is {sk_computed_smeared}, should be {sk_expected_smeared}"



import logging
from warnings import warn

from snewpy import get_models, model_path
from . import ccsn, presn


def __getattr__(name):
    if name in dir(ccsn):
        warn(f"{__name__}.{name} is moved to {__name__}.ccsn.{name}", FutureWarning, stacklevel=2)
        return getattr(ccsn, name)
    raise AttributeError(f"module {__name__} has no attribute {name}")


def _init_model(model_name, download=True, download_dir=model_path, **user_param):
    """Attempts to retrieve instantiated SNEWPY model using model class name and model parameters.
    If a model name is valid, but is not found and `download`=True, this function will attempt to download the model

    Parameters
    ----------
    model_name : str
        Name of SNEWPY model to import, must exactly match the name of the corresponding model class
    download : bool
        Switch for attempting to download model data if the first load attempt failed due to a missing file.
    download_dir : str
        Local directory to download model files to.
    user_param : varies
        User-requested model parameters used to initialize the model, if one is found.
        Error checking is performed during model initialization

    Raises
    ------
    ValueError
        If the requested model_name does not match any SNEWPY models

    See Also
    --------
    snewpy.models.ccsn
    snewpy.models.presn

    Example
    -------
    >>> from snewpy.models import _init_model; import astropy.units as u
    >>> _init_model('Nakazato_2013', progenitor_mass=13*u.Msun, metallicity=0.004, revival_time=0*u.s, eos='shen')
    Nakazato_2013 Model: nakazato-shen-BH-z0.004-s30.0.fits
    Progenitor mass  : 30.0 solMass
    EOS              : Shen
    Metallicity      : 0.004
    Revival time     : 0.0 ms

    :meta private:
    """
    if model_name in dir(ccsn):
        module = ccsn
    elif model_name in dir(presn):
        module = presn
    else:
        raise ValueError(f"Unable to find model with name '{model_name}' in snewpy.models.ccsn or snewpy.models.presn")

    try:
        return getattr(module, model_name)(**user_param)
    except FileNotFoundError as e:
        logger = logging.getLogger()
        logger.warning(f"Unable to find model {model_name} in {download_dir}")
        if not download:
            raise e
        logger.warning(f"Attempting to download model...")
        get_models(model_name, download_dir)
        return getattr(module, model_name)(**user_param)



# -*- coding: utf-8 -*-
"""Unit tests for the neutrino submodule.
"""
import unittest

from snewpy.neutrino import Flavor
from snewpy.models.loaders import Nakazato_2013, Tamborra_2014, OConnor_2013, OConnor_2015, \
                                  Sukhbold_2015, Bollig_2016, Walk_2018, \
                                  Walk_2019, Fornax_2019, Warren_2020, \
                                  Kuroda_2020, Fornax_2021, Zha_2021
from snewpy._model_urls import model_urls
from astropy import units as u
from snewpy import model_path
import os


class TestModels(unittest.TestCase):

    def test_model_urls(self):
        """Test that snewpy._model_urls.model_urls is empty. This should be populated if snewpy is downloaded from PyPI.
        This serves as a guard against accidentally committing/merging a populated model_urls to main.
        """
        self.assertFalse(model_urls)

    def test_Nakazato_2013(self):
        """
        Instantiate a set of 'Nakazato 2013' models
        """
        for z in [0.004, 0.02]:
            for trev in [100, 200, 300]:
                for mass in [13., 20., 50.]:
                    mfile = 'Nakazato_2013/nakazato-shen-z{}-t_rev{}ms-s{:.1f}.fits'.format(z, trev, mass)
                    model = Nakazato_2013(os.path.join(model_path, mfile))

                    # Check that times are in proper units.
                    t = model.get_time()
                    self.assertTrue(t.unit, u.s)
                    self.assertEqual(model.time[0], -50*u.ms)

                    # Check that we can compute flux dictionaries.
                    f = model.get_initial_spectra(0*u.s, 10*u.MeV)
                    self.assertEqual(type(f), dict)
                    self.assertEqual(len(f), len(Flavor))
                    self.assertEqual(f[Flavor.NU_E].unit, 1/(u.erg * u.s))

    def test_Tamborra_2014(self):
        """
        Instantiate a set of 'Tamborra 2014' models
        """
        for mass in [20., 27.]:
            for direction in [1, 2, 3]:
                mfile = 'Tamborra_2014/s{:.1f}c_3D_dir{}'.format(mass, direction)
                if mass == 20. and direction > 1:
                    with self.assertRaises(FileNotFoundError):
                        model = Tamborra_2014(os.path.join(model_path, mfile), eos='LS220')
                else:
                    model = Tamborra_2014(os.path.join(model_path, mfile), eos='LS220')

                    # Check that times are in proper units.
                    t = model.get_time()
                    self.assertTrue(t.unit, u.s)

                    # Check that we can compute flux dictionaries.
                    f = model.get_initial_spectra(0*u.s, 10*u.MeV)
                    self.assertEqual(type(f), dict)
                    self.assertEqual(len(f), len(Flavor))
                    self.assertEqual(f[Flavor.NU_E].unit, 1/(u.erg * u.s))

    def test_OConnor_2013(self):
        """
        Instantiate a set of "O'Connor 2015" models
        """
        for mass in list(range(12, 34)) + list(range(35, 61, 5)) + [70, 80, 100, 120]:
            for eos in ['LS220', 'HShen']:
                metadata = {
                    'Progenitor mass': mass * u.Msun,
                    'EOS': eos,
                }
                model = OConnor_2013(filename=f'{eos}_timeseries.tar.gz', metadata=metadata)

                self.assertEqual(model.metadata['EOS'], eos)
                self.assertEqual(model.metadata['Progenitor mass'].value, mass)

                # Check that times are in proper units.
                t = model.get_time()
                self.assertTrue(t.unit, u.s)

                # Check that we can compute flux dictionaries.
                f = model.get_initial_spectra(0*u.s, 10*u.MeV)
                self.assertEqual(type(f), dict)
                self.assertEqual(len(f), len(Flavor))
                self.assertEqual(f[Flavor.NU_E].unit, 1/(u.erg * u.s))

    def test_OConnor_2015(self):
        """
        Instantiate a set of "O'Connor 2015" models
        """
        mfile = 'M1_neutrinos.dat'
        model = OConnor_2015(mfile)

        # Check that times are in proper units.
        t = model.get_time()
        self.assertTrue(t.unit, u.s)

        # Check that we can compute flux dictionaries.
        f = model.get_initial_spectra(0*u.s, 10*u.MeV)
        self.assertEqual(type(f), dict)
        self.assertEqual(len(f), len(Flavor))
        self.assertEqual(f[Flavor.NU_E].unit, 1/(u.erg * u.s))

    def test_Sukhbold_2015(self):
        """
        Instantiate a set of 'Sukhbold 2015' models
        """
        for mass in ['z9.6', 's27.0']:
            for eos in ['LS220', 'SFHo']:
                mfile = 'Sukhbold_2015/sukhbold-{}-{}.fits'.format(eos, mass)
                model = Sukhbold_2015(os.path.join(model_path, mfile))

                # Check that times are in proper units.
                t = model.get_time()
                self.assertTrue(t.unit, u.s)

                # Check that we can compute flux dictionaries.
                f = model.get_initial_spectra(0*u.s, 10*u.MeV)
                self.assertEqual(type(f), dict)
                self.assertEqual(len(f), len(Flavor))
                self.assertEqual(f[Flavor.NU_E].unit, 1/(u.erg * u.s))

    def test_Bollig_2016(self):
        """
        Instantiate a set of 'Bollig 2016' models
        """
        for mass in [11.2, 27.]:
            mfile = 'Bollig_2016/s{:.1f}c'.format(mass)
            model = Bollig_2016(os.path.join(model_path, mfile), eos='LS220')

            # Check that times are in proper units.
            t = model.get_time()
            self.assertTrue(t.unit, u.s)

            # Check that we can compute flux dictionaries.
            f = model.get_initial_spectra(0*u.s, 10*u.MeV)
            self.assertEqual(type(f), dict)
            self.assertEqual(len(f), len(Flavor))
            self.assertEqual(f[Flavor.NU_E].unit, 1/(u.erg * u.s))

    def test_Walk_2018(self):
        """
        Instantiate a set of 'Walk 2018' models
        """
        mass = 15.
        for direction in [1,2,3]:
            for rot in ['fast','slow','non']:
                mfile = 'Walk_2018/s{:.1f}c_3D_{}rot_dir{}'.format(mass,rot,direction)
                model = Walk_2018(os.path.join(model_path, mfile), eos='LS220')

                # Check that times are in proper units.
                t = model.get_time()
                self.assertTrue(t.unit, u.s)

                # Check that we can compute flux dictionaries.
                f = model.get_initial_spectra(0*u.s, 10*u.MeV)
                self.assertEqual(type(f), dict)
                self.assertEqual(len(f), len(Flavor))
                self.assertEqual(f[Flavor.NU_E].unit, 1/(u.erg * u.s))

    def test_Walk_2019(self):
        """
        Instantiate a set of 'Walk 2019' models
        """
        for mass in [40.,75.]:
            for direction in [1,2]:
                mfile = 'Walk_2019/s{:.1f}c_3DBH_dir{}'.format(mass,direction)
                model = Walk_2019(os.path.join(model_path, mfile), eos='LS220')

                self.assertEqual(model.metadata['EOS'], 'LS220')
                self.assertEqual(model.metadata['Progenitor mass'], mass*u.Msun)

                # Check that times are in proper units.
                t = model.get_time()
                self.assertTrue(t.unit, u.s)

                # Check that we can compute flux dictionaries.
                f = model.get_initial_spectra(0*u.s, 10*u.MeV)
                self.assertEqual(type(f), dict)
                self.assertEqual(len(f), len(Flavor))
                self.assertEqual(f[Flavor.NU_E].unit, 1/(u.erg * u.s))

    def test_Fornax_2019(self):
        """
        Instantiate a set of 'Fornax 2019' models
        """
        for mass in [9, 10, 12, 13, 14, 15, 19, 25, 60]:
            mfile = 'Fornax_2019/lum_spec_{}M.h5'.format(mass)
            model = Fornax_2019(os.path.join(model_path, mfile), metadata={'Progenitor mass': mass*u.Msun})

            self.assertEqual(model.metadata['Progenitor mass'], mass*u.Msun)

            # Check that times are in proper units.
            t = model.get_time()
            self.assertTrue(t.unit, u.s)

            # Check that we can compute flux dictionaries.
            f = model.get_initial_spectra(0*u.s, 10*u.MeV, theta=23*u.degree, phi=22*u.degree)
            self.assertEqual(type(f), dict)
            self.assertEqual(len(f), len(Flavor))
            self.assertEqual(f[Flavor.NU_E].unit, u.erg/(u.MeV * u.s))

    def test_Warren_2020(self):
        """
        Instantiate a set of 'Warren 2020' models
        """
        masses = [
            '10.0', '10.25', '10.5', '10.75', '100', '11.0', '11.25', '11.5', '11.75',
            '12.0', '12.25', '12.5', '12.75', '120', '13.0', '13.1', '13.2', '13.3',
            '13.4', '13.5', '13.6', '13.7', '13.8', '13.9', '14.0', '14.1', '14.2',
            '14.3', '14.4', '14.5', '14.6', '14.7', '14.8', '14.9', '15.0', '15.1',
            '15.2', '15.3', '15.4', '15.5', '15.6', '15.7', '15.8', '15.9', '16.0',
            '16.1', '16.2', '16.3', '16.4', '16.5', '16.6', '16.7', '16.8', '16.9',
            '17.0', '17.1', '17.2', '17.3', '17.4', '17.5', '17.6', '17.7', '17.8',
            '17.9', '18.0', '18.1', '18.2', '18.3', '18.4', '18.5', '18.6', '18.7',
            '18.8', '18.9', '19.0', '19.1', '19.2', '19.3', '19.4', '19.5', '19.6',
            '19.7', '19.8', '19.9', '20.0', '20.1', '20.2', '20.3', '20.4', '20.5',
            '20.6', '20.7', '20.8', '20.9', '21.0', '21.1', '21.2', '21.3', '21.4',
            '21.5', '21.6', '21.7', '21.8', '21.9', '22.0', '22.1', '22.2', '22.3',
            '22.4', '22.5', '22.6', '22.7', '22.8', '22.9', '23.0', '23.1', '23.2',
            '23.3', '23.4', '23.5', '23.6', '23.7', '23.8', '23.9', '24.0', '24.1',
            '24.2', '24.3', '24.4', '24.5', '24.6', '24.7', '24.8', '24.9', '25.0',
            '25.1', '25.2', '25.3', '25.4', '25.5', '25.6', '25.7', '25.8', '25.9',
            '26.0', '26.1', '26.2', '26.3', '26.4', '26.5', '26.6', '26.7', '26.8',
            '26.9', '27.0', '27.1', '27.2', '27.3', '27.4', '27.5', '27.6', '27.7',
            '27.8', '27.9', '28.0', '28.1', '28.2', '28.3', '28.4', '28.5', '28.6',
            '28.7', '28.8', '28.9', '29.0', '29.1', '29.2', '29.3', '29.4', '29.5',
            '29.6', '29.7', '29.8', '29.9', '30.0', '31', '32', '33', '35', '40', '45',
            '50', '55', '60', '70', '80', '9.0', '9.25', '9.5', '9.75']

        for mixing in [1.23, 1.25, 1.27]:
            for mass in masses:
                mfile = f'stir_a{mixing}/stir_multimessenger_a{mixing}_m{mass}.h5'
                metadata = {
                    'Progenitor mass': float(mass) * u.Msun,
                    'Turb. mixing param.': mixing,
                    'EOS': 'SFHo',
                }
                model = Warren_2020(mfile, metadata)

                # Check that times are in proper units.
                t = model.get_time()
                self.assertTrue(t.unit, u.s)

                # Check that we can compute flux dictionaries.
                f = model.get_initial_spectra(0 * u.s, 10 * u.MeV)
                self.assertEqual(type(f), dict)
                self.assertEqual(len(f), len(Flavor))
                self.assertEqual(f[Flavor.NU_E].unit, 1. / (u.erg * u.s))

    def test_Kuroda_2020(self):
        """
        Instantiate a set of 'Kuroda 2020' models
        """
        for field in ['R00B00', 'R10B12', 'R10B13']:
            mfile = 'Kuroda_2020/Lnu{}.dat'.format(field)
            model = Kuroda_2020(os.path.join(model_path, mfile))

            # Check that times are in proper units.
            t = model.get_time()
            self.assertTrue(t.unit, u.s)

            # Check that we can compute flux dictionaries.
            f = model.get_initial_spectra(0*u.s, 10*u.MeV)
            self.assertEqual(type(f), dict)
            self.assertEqual(len(f), len(Flavor))
            self.assertEqual(f[Flavor.NU_E].unit, 1/(u.erg * u.s))

    def test_Fornax_2021(self):
        """
        Instantiate a set of 'Fornax 2021' models
        """
        for mass in ['12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22', '23', '25', '26', '26.99']:
            mfile = 'Fornax_2021/lum_spec_{}M_r10000_dat.h5'.format(mass)
            model = Fornax_2021(os.path.join(model_path, mfile), metadata={'Progenitor mass': mass*u.Msun})

            self.assertEqual(model.metadata['Progenitor mass'], float(mass)*u.Msun)

            # Check that times are in proper units.
            t = model.get_time()
            self.assertTrue(t.unit, u.s)

            # Check that we can compute flux dictionaries.
            f = model.get_initial_spectra(0*u.s, 10*u.MeV)
            self.assertEqual(type(f), dict)
            self.assertEqual(len(f), len(Flavor))
            self.assertEqual(f[Flavor.NU_E].unit, 1/(u.erg * u.s))

    def test_Zha_2021(self):
        """
        Instantiate a set of 'Zha 2021' models
        """
        for mass in ['16', '17', '18', '19', '19.89', '20', '21', '22.39', '23', '24', '25', '26', '30', '33']:
            mfile = 'Zha_2021/s{}.dat'.format(mass)
            model = Zha_2021(os.path.join(model_path, mfile))

            # Check that times are in proper units.
            t = model.get_time()
            self.assertTrue(t.unit, u.s)

            # Check that we can compute flux dictionaries.
            f = model.get_initial_spectra(0*u.s, 10*u.MeV)
            self.assertEqual(type(f), dict)
            self.assertEqual(len(f), len(Flavor))
            self.assertEqual(f[Flavor.NU_E].unit, 1./(u.erg * u.s))




import os
from warnings import warn

from snewpy.snowglobes import collate as NEW_collate

def collate(Branch, Model_Path, Tarball, detector_input="all", skip_plots=False, return_tables=False, verbose=False, remove_generated_files=True):
    # Deprecated since SNEWPY v1.1
    warn(f"The 'snewpy.from_snowglobes.collate()' function is deprecated. Use 'snewpy.snowglobes.collate()' instead.", FutureWarning)
    tarball_path = os.path.join(Model_Path, Tarball)
    return NEW_collate(Branch, tarball_path, skip_plots=skip_plots)



# -*- coding: utf-8 -*-
"""This module implements basic neutrino properties that are used throughout SNEWPY."""

from enum import IntEnum
from astropy import units as u


class MassHierarchy(IntEnum):
    """Neutrino mass ordering: ``NORMAL`` or ``INVERTED``."""
    NORMAL = 1
    INVERTED = 2


class Flavor(IntEnum):
    """Enumeration of CCSN Neutrino flavors."""
    NU_E = 0
    NU_X = 1
    NU_E_BAR = 2
    NU_X_BAR = 3
    
    def to_tex(self):
        """LaTeX-compatible string representations of flavor."""
        if '_BAR' in self.name:
            return r'$\overline{{\nu}}_{0}$'.format(self.name[3].lower())
        return r'$\{0}$'.format(self.name.lower())

    @property
    def is_electron(self):
        """Return ``True`` for ``Flavor.NU_E`` and ``Flavor.NU_E_BAR``."""
        return self.value in (Flavor.NU_E.value, Flavor.NU_E_BAR.value)

    @property
    def is_neutrino(self):
        """Return ``True`` for ``Flavor.NU_E`` and ``Flavor.NU_X``."""
        return self.value in (Flavor.NU_E.value, Flavor.NU_X.value)

    @property
    def is_antineutrino(self):
        return self.value in (Flavor.NU_E_BAR.value, Flavor.NU_X_BAR.value)


class MixingParameters:
    """Mixing angles and mass differences, assuming three neutrino flavors.
    
    This class contains the default values used throughout SNEWPY, which are
    based on `NuFIT 5.0 <http://www.nu-fit.org>`_ results from July 2020,
    published in `JHEP 09 (2020) 178 <https://dx.doi.org/10.1007/JHEP09(2020)178>`_
    [`arXiv:2007.14792 <https://arxiv.org/abs/2007.14792>`_].
    Note that the best fit values vary between normal and inverted mass hierarchy.
    """
    def __init__(self, mh=MassHierarchy.NORMAL):
        """Initialize the neutrino mixing parameters.

        Parameters
        ----------
        mh : MassHierarchy
            Desired mass ordering: NORMAL or INVERTED.
        """
        if type(mh) == MassHierarchy:
            self.mass_order = mh
        else:
            raise TypeError('mh must be of type MassHierarchy')

        # Values from JHEP 09 (2020) 178 [arXiv:2007.14792] and www.nu-fit.org.
        # The reported precision is not significant given current
        # uncertainties, but is useful for comparing to the table of
        # parameters presented on nu-fit.org.
        if self.mass_order == MassHierarchy.NORMAL:
            # Note: in NH, the mass splittings are: m1..m2..............m3.
            self.theta12 = 33.44 * u.deg
            self.theta13 =  8.57 * u.deg
            self.theta23 = 49.20 * u.deg
            self.deltaCP = 197 * u.deg
            self.dm21_2  = 7.42e-5 * u.eV**2
            self.dm32_2  = 2.517e-3 * u.eV**2
        else:
            # Note: in IH, the mass splittings are: m3..............m1..m2.
            self.theta12 = 33.45 * u.deg
            self.theta13 =  8.60 * u.deg
            self.theta23 = 49.30 * u.deg
            self.deltaCP = 282 * u.deg
            self.dm21_2  = 7.42e-5 * u.eV**2
            self.dm31_2  = -2.498e-3 * u.eV**2

    def get_mixing_angles(self):
        """Mixing angles of the PMNS matrix.
        
        Returns
        -------
        tuple
            Angles theta12, theta13, theta23.
        """
        return (self.theta12, self.theta13, self.theta23)




#!/usr/bin/env python
#
# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
from distutils.command.sdist import sdist as DistutilsSdist
from setuptools import setup, find_packages

# Git-based version info. Remove?
from python.snewpy._git import get_version, SetVersion

#
# Begin setup
#
setup_keywords = dict()
#
setup_keywords['name'] = 'snewpy'
setup_keywords['description'] = 'A Python package for working with supernova neutrinos'
setup_keywords['author'] = 'SNEWS Collaboration'
setup_keywords['author_email'] = 'snews2.0@lists.bnl.gov'
setup_keywords['license'] = 'BSD'
setup_keywords['url'] = 'https://github.com/SNEWS2/snewpy'
setup_keywords['version'] = get_version()
#
# Use README.md as a long_description.
#
setup_keywords['long_description'] = ''
if os.path.exists('README.md'):
    with open('README.md') as readme:
        setup_keywords['long_description'] = readme.read()
    setup_keywords['long_description_content_type'] = 'text/markdown'
#
# Set other keywords for the setup function.
#
# Use entry_points to let `pip` create executable scripts for each target platform.
# See https://setuptools.readthedocs.io/en/latest/userguide/entry_point.html
# setup_keywords['entry_points'] = {'console_scripts': ['to_snowglobes = snewpy.to_snowglobes:generate_time_series', ], },
setup_keywords['provides'] = [setup_keywords['name']]
setup_keywords['python_requires'] = '>=3.8'
setup_keywords['zip_safe'] = False
setup_keywords['packages'] = find_packages('python')
setup_keywords['package_dir'] = {'': 'python'}
setup_keywords['package_data'] = {'':['templates/*.glb', 'models/*.yml']}
setup_keywords['cmdclass'] = {'version': SetVersion, 'sdist': DistutilsSdist}
setup_keywords['test_suite']='snewpy.test.snewpy_test_suite.snewpy_test_suite'

requires = []
with open('requirements.txt', 'r') as f:
    for line in f:
        if line.strip():
            requires.append(line.strip())
setup_keywords['install_requires'] = requires
setup_keywords['extras_require'] = {  # Optional
    'dev': ['pytest'],
    'docs':['numpydoc']
}
#
# Internal data directories.
#
#setup_keywords['data_files'] = [('snewpy/data/config', glob('data/config/*')),
#                                ('snewpy/data/spectra', glob('data/spectra/*'))]
#
# Run setup command.
#
setup(**setup_keywords)
